package pmp;

import java.util.*;
import java.util.regex.*;
import static pmp.ConfigDirectives.*;

/**
 * Syntactic analyzer.
 * <p>
 * Title: PMP: Macroprocessor
 * </p>
 * <p>
 * Description: Java macroprocessor
 * </p>
 * <p>
 * Copyright: Copyright (c) 2005
 * </p>
 * 
 * @author Lud\u011Bk Hlav\u00E1\u010Dek
 * @version 1.0
 */

public final class Analyzer extends Macroprocessor.Component implements
        MacroListModifyListener {

    /**
     * Constructs lexical analyzer.
     * 
     * @param mp
     *            owner of this object
     */
    public Analyzer(Macroprocessor mp)
    {
        mp.super();

        regexps = new EnumMap<State, RegExps>(State.class);
        for (State s : State.values()) {
            regexps.put(s, new RegExps(mp));
        } // for
        strregexps = new Hashtable<String, RegExps>();
    }

    /** Array of RegExps indexed by states */
    private final EnumMap<State, RegExps> regexps;

    /** Map of RegExps objects for strings; keys are string ID's */
    private final Map<String, RegExps> strregexps;

    /** States of syntactic analyzer. */
    public enum State {
        START, PLAINTEXT, CODE_BLOCK, STRING, MACRO, PARAMS, EOF
    }

    /** Current state of lexical analyzer. */
    protected State synState;

    /**
     * Returns state of parser as string
     * 
     * @return String describing state of analyzer
     */
    public String stateAsString()
    {
        return stateAsString(synState);
    }

    /**
     * Converts integer representation of parser state to string
     * 
     * @param state
     *            state of parser
     * @return string representation of state
     */
    public String stateAsString(State state)
    {
        return "STATE_" + state.toString();
    }

    /**
     * Returns string representation of this object. Returned String consists of
     * result of <code>toString()</code> method of superclass and state of
     * parser.
     * 
     * @return string representation of this class
     */
    @Override
    public String toString()
    {
        return super.toString() + " [" + stateAsString() + "]";
    }

    /**
     * Macro call stack.
     */
    protected final CallStack callStack = new CallStack();

    /**
     * Build list of regexps for given state.
     * 
     * @param state
     *            parser state
     * @param re
     *            regexps buffer
     */
    private void buildRegexpForState(State state)
    {

        RegExps re = regexps.get(state);
        re.clear();

        if (state == State.PLAINTEXT) {
            re.addRegExp(macros.getString(CODE_START), // CODE START
                    new Token(Token.Type.CODE_START));
        }
        if ((state == State.CODE_BLOCK) || (state == State.PARAMS)) { // STRINGS
            String[] stringIDs = macros.getIDsByMask(STRING__START);
            for (int i = 0; i < stringIDs.length; i++) {
                if ((state == State.PARAMS)
                        || macros.getBool(STRING__STANDALONE, stringIDs[i])) {
                    re.addRegExp(macros.getString(STRING__START, stringIDs[i]),
                            new Token(Token.Type.STRING_START, stringIDs[i]));
                }
            } // for
        }
        if (state == State.MACRO) {
            re.addRegExp(macros.getString(MACRO_OPENING_BRACKET), new Token(
                    Token.Type.OPENING_BRACKET)); // OPENING BRACKET
        }

        if (state == State.PARAMS) {
            re.addRegExp(macros.getString(MACRO_PARAM_SEPARATOR), new Token(
                    Token.Type.PARAM_SEPARATOR)); // PARAM SEPARATOR
        }

        if ((state == State.CODE_BLOCK) || (state == State.PARAMS)) {
            re.addRegExp(macros.getString(MACRO_PATTERN), new Token(
                    Token.Type.MACRO_NAME)); // MACRO PATTERN
        }

        if (state != State.STRING) {
            re.addRegExp(macros.getString(PARSER_TOKEN), // TOKEN
                    new Token((state == State.PLAINTEXT) ? Token.Type.PLAINTEXT
                            : Token.Type.CODE));
        }

        if (state == State.PARAMS) {
            re.addRegExp(macros.getString(MACRO_CLOSING_BRACKET), new Token(
                    Token.Type.CLOSING_BRACKET)); // CLOSING BRACKET
        }

        if (state == State.CODE_BLOCK) {
            String codeend = macros.getString(CODE_END);
            if (!"".equals(codeend)) {
                re.addRegExp(codeend, new Token(Token.Type.CODE_END)); // CODE
                // END
            }
        }

        /*
         * if (state == State.STRING) { // SINGLECHAR
         * re.addRegExp(SINGLE_CHAR_TOKEN, new Token(Token.STRING_TOKEN)); }
         * else if (state == State.PLAINTEXT) { re.addRegExp(SINGLE_CHAR_TOKEN,
         * new Token(Token.PLAINTEXT)); } else if ((state == STATE_CODE_BLOCK) ||
         * (state == State.PARAMS)) { re.addRegExp(SINGLE_CHAR_TOKEN, new
         * Token(Token.CODE)); }
         */

        // System.err.println("For state: " + stateAsString(state));
        // System.err.println(re.toString());
    }

    /**
     * Creates regexps lists for string tokens.
     */
    private void buildRegexpForStrings()
    {

        MacroList ml = getMPMacros();

        String[] ids = ml.getIDsByMask(STRING__START);
        strregexps.clear();

        for (int i = 0; i < ids.length; i++) {
            RegExps re = new RegExps(getMP());

            String end;
            if (ml.isDirectiveDefined(STRING__END, ids[i])) {
                end = ml.getString(STRING__END, ids[i]); // END-OF-STRING
            }
            else {
                end = ml.getString(STRING__START, ids[i]);
                // same as START-OF-STRING
                // TODO: check this
                // notice("Missing definition of end-of-string symbol for string
                // type " + ids[i]);
            }
            re.addRegExp(end, new Token(Token.Type.STRING_END));

            String mode = ml.getStringL(STRING__MODE, ids[i]); // SPECIAL-SYMBOL

            if (mode.equals(VALUE_NESTED)) {
                String start = ml.getString(STRING__START, ids[i]);
                re.addRegExp(start, new Token(Token.Type.STRING_START));
            }

            if (mode.equals(VALUE_PARSE)) {
                String symbolPattern = ml.getString(STRING__SYMBOLS, ids[i]);
                if (symbolPattern != null) {
                    re.addRegExp(symbolPattern, new Token(
                            Token.Type.STRING_SPECIAL_SYMBOL));
                }
                else {
                    warning("Missing definition of special symbols for string type "
                            + ids[i]);
                }
            }

            if (ml.isDirectiveDefined(STRING__TOKENS, ids[i])) { // STRING-TOKEN
                String token = ml.getString(STRING__TOKENS, ids[i]);
                re.addRegExp(token, new Token(Token.Type.STRING_TOKEN));
            }

            strregexps.put(ids[i], re);
        } // for

    }

    /**
     * Listener method for macro list modifications.
     * 
     * @param event
     *            event object describing modification
     */
    public void macroListModified(MacroListModifyEvent event)
    {
        EnumSet<State> dirty = EnumSet.noneOf(State.class);

        if (event.isFlagSet(MacroListModifyEvent.MASK_PMP_PREFIX)) {
            dirty = EnumSet.allOf(State.class);
        }

        if (event.isFlagSet(MacroListModifyEvent.MASK_PMP_MACRO)) {
            dirty.add(State.MACRO);
            dirty.add(State.PARAMS);
            dirty.add(State.CODE_BLOCK);
        }
        if (event.isFlagSet(MacroListModifyEvent.MASK_PMP_STRING)) {
            dirty.add(State.CODE_BLOCK);
            dirty.add(State.PARAMS);
        }
        if (event.isFlagSet(MacroListModifyEvent.MASK_PMP_CODE)) {
            dirty.add(State.PLAINTEXT);
            dirty.add(State.CODE_BLOCK);
            dirty.add(State.PARAMS);
        }
        /*
         * if (event.isFlagSet(event.MASK_PMP_SYMBOLS)) { dirty[State.STRING] =
         * true; }
         */

        for (State dstate : dirty) {
            buildRegexpForState(dstate);
        } // for

        if (event.isFlagSet(MacroListModifyEvent.MASK_PMP_STRING) // rebuild
                // strings
                || (event.isFlagSet(MacroListModifyEvent.MASK_PMP_PREFIX))) {
            buildRegexpForStrings();
        }
    }

    /**
     * Performs macro execution.
     * <p>
     * Note: called only by {@link #processInput()}
     * </p>
     * 
     * @param prefetch
     *            token that should not have been read and will be returned to
     *            input queue.
     * @see Macroprocessor#runMacro(CallStackItem)
     */
    private void execMacroFromStack(String prefetch)
    {
        CallStackItem item = callStack.popItem();
        mp.debug(item, Macroprocessor.DEBUG_MACRO_CALLS);

        String result = getMP().runMacro(item);
        mp
                .debug("[RESULT: " + result + "]",
                        Macroprocessor.DEBUG_MACRO_RESULTS);

        if (prefetch != null) {
            iq.pushFront(prefetch); // string to be returned to IQ
        }
        iq.pushFront(result);
        if (callStack.topItemType() == CallStackItem.Type.MACRO) {
            synState = State.PARAMS;
        }
        else {
            synState = State.CODE_BLOCK;
        } // if/else

    }

    /**
     * Tests if string in token represents existing macro and creates new
     * callStack item.
     * <p>
     * Note: called only by {@link #processInput()}.
     * </p>
     * 
     * @param token
     *            with macro name string
     */
    private void checkForMacroName(Token token)
    {
        AbstractMacro macro = null;

        String pattern = macros.getString(MACRO_PATTERN);
        Pattern p = RegExps.compilePattern(pattern);
        Matcher m = p.matcher(token.getString());
        String macroName;
        if (m.matches()) {
            switch (m.groupCount()) {
            case 0:
                macroName = m.group();
                break;
            case 1:
                macroName = m.group(1);
                break;
            default:
                StringBuilder b = new StringBuilder(token.getString().length());
                for (int i = 1; i <= m.groupCount(); i++)
                    if (m.group(i) != null) b.append(m.group(i));
                macroName = b.toString();
                break;
            }
            macro = macros.getMacro(macroName);

            // try case-insensitive search
            if ((macro == null) && (!macros.getBool(MACRO_CASE_SENSITIVE))) {
                macro = macros.getMacroCI(macroName); // try case-insensitive
                // search
            } // if

            // try fall-back macro
            if (macro == null) {
                macro = macros.getMacro(macros.forName(MACRO_FALLBACK));
            }

        } // if matches

        if (macro != null) {
            callStack.pushMacroItem(token.getString(), macro);
            synState = State.MACRO;
        } // if
        else {
            if (synState == State.PARAMS) {
                callStack.appendToLastField(token.getString());
            }
            else {
                getMPIO().write(token.getString());
            }//else
        }//else
    }

    /**
     * Resets analyzer.
     * <p>
     * Note: called only by {@link #processInput()}.
     * </p>
     */
    private void reset()
    {
        macros = getMPMacros();
        output = getMPIO();
        iq = getMPIQ();
        mp = getMP();

        synState = State.START;
        callStack.clear();

        macros.addModifyListener(this);
    }

    /**
     * Parses string.
     * <p>
     * Note: called only by {@link #processInput()}
     * </p>
     * 
     * @param starttoken
     *            token with start-of-string symbol
     * @return processed string
     */
    private String stringParser(Token starttoken)
    {
        final int MODE_DISCARD = 57801;
        final int MODE_COPY = 57802;
        final int MODE_PARSE = 57803;
        final int MODE_NESTED = 57804;

        int mode = MODE_COPY;
        int strLevel = 1;
        StringBuilder result = new StringBuilder(128);

        boolean outputEndMark = false;
        String marksmode = macros.getStringL(STRING__MARKS_MODE, starttoken
                .getTypeID());
        if (VALUE_COPY.equals(marksmode)) {
            outputEndMark = true;
            result.append(starttoken.getString());
        } // if
        else if (VALUE_COPY_START.equals(marksmode)) {
            result.append(starttoken.getString());
        }
        else if (VALUE_COPY_END.equals(marksmode)) {
            outputEndMark = true;
        }

        String strmode = macros
                .getStringL(STRING__MODE, starttoken.getTypeID());
        AbstractMacro parser = null;
        String parserName = macros.forName(STRING__PARSER, starttoken
                .getTypeID());

        if (VALUE_PARSE.equals(strmode)) { // load parser
            mode = MODE_PARSE;
            parser = macros.getMacro(parserName);
            if (parser == null) {
                mode = MODE_COPY;
                warning("Missing parser definition for string type "
                        + starttoken.getTypeID());
            } // if
        }
        else if (VALUE_DISCARD.equals(strmode)) {
            mode = MODE_DISCARD;
        }
        else if (VALUE_NESTED.equals(strmode)) {
            mode = MODE_NESTED;
        }
        else if (VALUE_COPY.equals(strmode)) {
            mode = MODE_COPY; // simply copy content to output
        }

        RegExps re = strregexps.get(starttoken.getTypeID());
        assert (re != null);

        while (true) {
            Token token = re.outputToken();

            mp.debug(token, Macroprocessor.DEBUG_EVERY_TOKEN);

            switch (token.getType()) {

            case STRING_END: // STRING_END
                mp.debug("[STRING: " + starttoken.getTypeID() + "]: "
                        + result.toString(), Macroprocessor.DEBUG_MACRO_CALLS);

                strLevel--;

                if (mode == MODE_DISCARD)
                    result.append(macros.getString(STRING__OUTPUT, starttoken
                            .getTypeID()));

                if ((mode != MODE_NESTED) || (strLevel == 0)) {
                    if (outputEndMark) {
                        result.append(token.getString());
                    } // if
                    return result.toString();
                }

                result.append(token.getString());
                break;

            case STRING_START: // STRING_START
                assert (mode == MODE_NESTED);
                strLevel++;
                result.append(token.getString());
                break;

            case STRING_SPECIAL_SYMBOL: // SPECIAL_SYMBOL
                assert (mode == MODE_PARSE);
                String[] args = new String[] { parserName, token.getString() };
                CallStackItem it = new CallStackItem(CallStackItem.Type.PARSER,
                        args, parser);
                result.append(mp.runMacro(it));
                break;

            case STRING_TOKEN: // STRING TOKEN
            case SINGLE_CHAR:
                if (mode != MODE_DISCARD) {
                    result.append(token.getString());
                }
                break;

            case END: // END
                error("Unexpected end of file");
                break;

            default:
                // unexpected token
                throw new InternalMPError(InternalMPError.Source.ANALYZER,0x529);
            } // switch

        } // while
    }

    /** Used by processInput and subsequently called private methods */
    private MacroList macros;

    /** Used by processInput and subsequently called private methods */
    private InputOutput output;

    /** Used by processInput and subsequently called private methods */
    private InputQueue iq;

    /** Used by processInput and subsequently called private methods */
    private Macroprocessor mp;

    /**
     * Executes program. Reads input, splits in into tokens and substitues macro
     * calls.
     * 
     * @throws this
     *             exception shouldn't be thrown as Writer should report errors
     *             to InputOutput object.
     */
    public void processInput()
    {
        reset();

        Token token = null;

        for (State dstate : State.values()) {
            buildRegexpForState(dstate);
        } // for
        buildRegexpForStrings();

        iq.pushFront(macros.getString(SOURCE_PREPEND));
        iq.pushBack(macros.getString(SOURCE_APPEND));

        while (true) {

            switch (synState) {

            case PLAINTEXT: // . PLAINTEXT
                if ((token.getType() == Token.Type.PLAINTEXT)
                        || (token.getType() == Token.Type.SINGLE_CHAR)) {
                    output.write(token.getString());
                }
                else if (token.getType() == Token.Type.CODE_START) {
                    if (macros.getBool(CODE_MARKS_OUTPUT)) {
                        output.write(token.getString());
                    } // if
                    synState = State.CODE_BLOCK;
                }
                else if (token.getType() == Token.Type.END) {
                    synState = State.EOF;
                }
                else {
                    throw new InternalMPError(0x10426); /* something is wrong */
                }
                break;

            case CODE_BLOCK: // CODE_BLOCK

                if (token.getType() == Token.Type.MACRO_NAME) {
                    checkForMacroName(token);
                }

                else if ((token.getType() == Token.Type.CODE)
                        || (token.getType() == Token.Type.SINGLE_CHAR)) {
                    getMPIO().write(token.getString());
                }

                else if (token.getType() == Token.Type.STRING_START) {
                    synState = State.STRING;
                    String str = stringParser(token);
                    output.write(str);
                    synState = State.CODE_BLOCK;
                }

                else if (token.getType() == Token.Type.CODE_END) {
                    if (macros.getBool(CODE_MARKS_OUTPUT)) {
                        output.write(token.getString());
                    } // if
                    synState = State.PLAINTEXT;
                }

                else if (token.getType() == Token.Type.END) {
                    synState = State.EOF;
                }

                else {
                    throw new InternalMPError(0x10463); /* unexpected token type */
                }
                break;

            case MACRO: // MACRO
                if (token.getType() == Token.Type.OPENING_BRACKET) {
                    synState = State.PARAMS;
                    callStack.topItem().openingBracket = token.getString();
                } // if
                else { // exec macro ?
                    if (macros.getBool(MACRO_ARGUMENTS_REQUIRED)) {
                        // no parameters were given -> ignore this macroname
                        iq.pushFront(token.getString()); // return this token
                        iq.pushFront(callStack.popItem().getParam0()); // return
                        // macro name
                        // token
                        if (callStack.topItemType() == CallStackItem.Type.MACRO) {
                            synState = State.PARAMS;
                        }
                        else {
                            synState = State.CODE_BLOCK;
                        } // else

                    } // if
                    else { // => parameters are not required for macro
                        // recognition
                        callStack.stripParamFields();
                        execMacroFromStack(token.getString());
                    } // else
                } // else
                break;

            case PARAMS: // PARAMS
                if (token.getType() == Token.Type.PARAM_SEPARATOR) {
                    callStack.addField("");
                }

                else if (token.getType() == Token.Type.CLOSING_BRACKET) {
                    callStack.topItem().closingBracket = token.getString();
                    execMacroFromStack(null);
                }

                else if (token.getType() == Token.Type.STRING_START) {
                    synState = State.STRING;
                    String str = stringParser(token);
                    callStack.appendToLastField(str);
                    synState = State.PARAMS;
                }

                else if (token.getType() == Token.Type.MACRO_NAME) {
                    checkForMacroName(token);
                }

                else if ((token.getType() == Token.Type.CODE)
                        || (token.getType() == Token.Type.SINGLE_CHAR)) {
                    callStack.appendToLastField(token.getString());
                }

                else if (token.getType() == Token.Type.END) {
                    synState = State.EOF;
                }

                else {
                    throw new InternalMPError(0x10515);
                }
                break;

            case STRING: // STRING
                throw new InternalMPError(0x10520);
            /* should not happen as string handling is done in stringParser() */

            case START:

                /** @todo something to be done on start */
                synState = State.PLAINTEXT;
                break;

            case EOF:
                if (!callStack.isEmpty()) {
                    error("Unexpected end of file");
                }
                return;
            } // switch

            token = regexps.get(synState).outputToken();
            mp.debug(token, Macroprocessor.DEBUG_EVERY_TOKEN);

        } // while
    }

}
