package pmp;

import java.util.LinkedList;
import java.util.List;

/**
 * Stack for storing macro parameters.
 * <p>
 * Title: PMP: Macroprocessor
 * </p>
 * <p>
 * Description: Java macroprocessor
 * </p>
 * <p>
 * Copyright: Copyright (c) 2005
 * </p>
 * 
 * @author Lud\u011Bk Hlav\u00E1\u010Dek
 * @version 1.0
 */

public final class CallStack {

    /**
     * Internal representation of the stack.
     */
    private final LinkedList<CallStackItem> stack = new LinkedList<CallStackItem>();

    /**
     * Pushes new item to stack.
     * 
     * @param type
     *            type of item
     * @param params
     *            parameters
     * @return newly createm CallStackItem object
     */
    public CallStackItem pushItem(CallStackItem.Type type, String[] params)
    {
        CallStackItem item = new CallStackItem(type, params);
        return pushItem(item);
    }

    /**
     * Pushes new item to stack. This item has type field set to
     * <code>TYPE_MACRO</code> and parameters are
     * <code>{ macroName, ""}</code>.
     * 
     * @param macroName
     *            name of macro as appeared in program
     * @param macro
     *            that will be invoked when all paameters are read
     * @return newly createm CallStackItem object
     */
    public CallStackItem pushMacroItem(String macroName, AbstractMacro macro)
    {
        String[] params = { macroName, "" };
        CallStackItem item = new CallStackItem(CallStackItem.Type.MACRO,
                params, macro);
        return pushItem(item);
    }

    /**
     * Pushes new item to stack.
     * 
     * @param item
     *            item to be added to the top of the stack.
     * @return <code>item</code>
     */
    public CallStackItem pushItem(CallStackItem item)
    {
        stack.addFirst(item);
        return item;
    }

    /**
     * Pops item from stack.
     * 
     * @return item removed from the top of the stack
     */
    public CallStackItem popItem()
    {
        return stack.removeFirst();
    }

    /**
     * Returns <code>true</code> if stack is empty.
     * 
     * @return <code>true</code> if empty
     */
    public boolean isEmpty()
    {
        return stack.isEmpty();
    }

    /**
     * Removes all items from stack.
     */
    public void clear()
    {
        stack.clear();
    }

    /**
     * Returns type of item on top of the stack. If stack is empty returns
     * CallStackItem.TYPE_EMPTY.
     * 
     * @return type of item
     * @see CallStackItem
     */
    public CallStackItem.Type topItemType()
    {
        if (stack.isEmpty()) { return CallStackItem.Type.EMPTY; }
        CallStackItem top = stack.getFirst();
        return top.getType();
    }

    /**
     * Returns item on top of the stack or <code>null</code> if stack is
     * empty.
     * 
     * @return top item from the stack
     */
    public CallStackItem topItem()
    {
        if (stack.isEmpty()) { return null; }
        return stack.getFirst();
    }

    /**
     * Adds new field into array on top of the stack.
     * 
     * @param str
     *            value
     * @return item on top of the stack
     */
    public CallStackItem addField(String str)
    {
        CallStackItem it = stack.getFirst();
        it.params.add(str);
        return it;
    }

    /**
     * Removes all parameter fields from item on top of the stack. This method
     * is used when macro is invoked without parameters.
     * 
     * @return item on top of the stack
     */
    public CallStackItem stripParamFields()
    {
        CallStackItem it = stack.getFirst();
        it.params.subList(1, it.params.size()).clear();
        return it;
    }

    /**
     * Appends <code>str</code>to last field in array on top of the stack.
     * 
     * @param str
     *            value
     * @return item on top of the stack
     */
    public CallStackItem appendToLastField(String str)
    {
        CallStackItem it = stack.getFirst();
        List<String> args = it.params;
        int last = args.size() - 1;
        args.set(last, args.get(last) + str);
        return it;
    }

}
