package pmp;

import java.util.*;

/**
 * Class representing item of call-stack.
 * <p>
 * Title: PMP: Macroprocessor
 * </p>
 * <p>
 * Description: Java macroprocessor
 * </p>
 * <p>
 * Copyright: Copyright (c) 2005
 * </p>
 * 
 * @author Lud\u011Bk Hlav\u00E1\u010Dek
 * @version 1.0
 */

public final class CallStackItem {

    public enum Type {
        EMPTY, MACRO, PARSER
    }

    private final Type type;

    private final AbstractMacro macro;

    /** List of arguments. Items are strings. */
    public final List<String> params;

    public String openingBracket;

    public String closingBracket;

    public CallStackItem(Type type, String[] params, AbstractMacro macro)
    {
        this.type = type;
        this.params = new ArrayList<String>(8);
        this.params.addAll(Arrays.asList(params));
        this.macro = macro;
    }

    public CallStackItem(Type type, String[] params)
    {
        this(type, params, null);
    }

    public Type getType()
    {
        return type;
    }

    public AbstractMacro getMacro()
    {
        return macro;
    }

    public String[] getStrParams()
    {
        return params.toArray(new String[0]);
    }

    /**
     * Returns first paramater (index 0).
     * 
     * @return first parameter
     */
    public String getParam0()
    {
        return params.get(0);
    }

    /**
     * Returns paramater at given index.
     * 
     * @return parameter at given index
     */
    public String getParam(int index)
    {
        return params.get(index);
    }

    /**
     * Returns number of actual parameters.
     * 
     * @return number of parameters
     */
    public int getParamCount()
    {
        return params.size();
    }

    /**
     * Adds empty strings into parameter list until it has enough items to
     * safely invoke macro.
     * 
     * @return number of parameters actually appended
     * @see pmp.AbstractMacro#getRequiredParams()
     */
    public int appendRequiredParams()
    {
        int paramsReq = macro.getRequiredParams();
        int added = 0;
        while (paramsReq >= params.size()) {
            // make sure there is at least paramsReq+1 parameters
            params.add("");
            added++;
        } // while
        return added;
    }

    /**
     * Returns string representation of this object.
     * 
     * @return string representation
     */
    @Override
    public String toString()
    {
        StringBuilder str = new StringBuilder(128);
        str.append("[").append(type.toString());
        if (macro != null) {
            str.append(": ").append(macro.getClass().getName());
        } // if
        str.append("]: (");
        if (params.size() > 0) {
            str.append("\"").append(params.get(0)).append("\"");
        } // if

        for (int i = 1; i < params.size(); i++) {
            str.append(", \"").append(params.get(i)).append("\"");
        } // for
        str.append(")");

        return str.toString();
    }
}
