package pmp;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import static pmp.ConfigDirectives.*;

/**
 * User-defined macro. It is used for configuration directives and other data
 * stored in macroprocessor's MacroList.
 * <p>
 * <b>Notes:</b> All descendants of AbstractMacro class (including this class)
 * are considered to be immutable. If you want to redefine macro, create new
 * object and change reference in {@link pmp.MacroList} using
 * {@link pmp.MacroList#setMacro(String, AbstractMacro)} or similar method.
 * </p>
 * <p>
 * Title: PMP: Macroprocessor
 * </p>
 * <p>
 * Description: Java macroprocessor
 * </p>
 * <p>
 * Copyright: Copyright (c) 2005
 * </p>
 * 
 * @author Lud\u011Bk Hlav\u00E1\u010Dek
 * @version 1.0
 */

public final class DefinedMacro extends AbstractMacro {

    /** expansion string */
    protected final String expansion;

    public DefinedMacro(String expansion)
    {
        this.expansion = expansion;
    }

    private static Pattern p = Pattern.compile("([0-9]+)(?:-([0-9]*))?");

    private class Range {

        public static final int MACRONAME = -1;

        public static final int REMAINING_PARAMS = -1;

        final int start;

        final int end;

        final int endpos;

        final String content;

        Range(int pos, char term)
        {
            endpos = expansion.indexOf(term, pos);
            String val = expansion.substring(pos, endpos);

            if (val.startsWith("\\")) {
                content = val.substring(1);
                start = MACRONAME;
                end = 0; // unused value
                return;
            }
            else {
                content = val;
            }
            if ("*".equals(val)) {
                start = 0;
                end = REMAINING_PARAMS;
                return;
            }
            Matcher m = p.matcher(val);
            if (m.matches()) {
                start = Integer.parseInt(m.group(1));
                if (m.group(2).length() > 0)
                    end = Integer.parseInt(m.group(2));
                else
                    end = REMAINING_PARAMS;
                return;
            }// if
            throw new FatalParserException("Invalid defined macro");
        }
    }// class Range

    private String getMacro(String macroName, Macroprocessor mp)
    {
        AbstractMacro m = mp.getMPMacros().getMacro(macroName);
        if (m != null) return m.toString();

        mp.warning("Defined macro references undefined macro " + macroName);
        return "";
    }

    /**
     * Expands given macro with given arguments.
     * 
     * @param mp
     *            {@inheritDoc}
     * @param args
     *            {@inheritDoc}
     * @return {@inheritDoc}
     */
    @Override
    public String run(Macroprocessor mp, String args[])
    {
        StringBuilder result = new StringBuilder(expansion.length() * 2);
        int i = 0;

        MacroList ml = mp.getMPMacros();
        String sep = ", ";
        if (ml.isDirectiveDefined(DEFINED_SEPARATOR))
            ml.getString(DEFINED_SEPARATOR);

        AbstractMacro encoder = ml.getMacro(ml.forName(DEFINED_SEPARATOR));

        while (i < expansion.length()) {

            char c = expansion.charAt(i++);
            if (c != '$') {
                result.append(c);
                continue;
            }

            c = expansion.charAt(i++);

            if (c == '$') { // char $
                result.append('$');
            }

            else if (c == '#') { // number of args
                result.append(args.length);
            }

            else if (c == '{') { // raw arguments
                Range r = new Range(i, '}');
                if (r.start == Range.MACRONAME)
                    result.append(getMacro(r.content, mp));
                else {
                    int j = r.start;
                    while (j != r.end) {
                        if (j >= args.length) break;
                        result.append(args[j]);
                        if (j < r.end) result.append(", ");
                        j++;
                    }// while
                }// else
                i = r.endpos + 1;
            }

            else if (c == '(') { // cooked arguments
                Range r = new Range(i, ')');
                if (r.start == Range.MACRONAME) {
                    String m = getMacro(r.content, mp);
                    if (encoder != null)
                        m = encoder.run(mp, new String[] { args[0], m });
                    result.append(m);
                }
                else {
                    int j = r.start;
                    while (j != r.end) {
                        if (j >= args.length) break;
                        String m = args[j];
                        if (encoder != null)
                            m = encoder.run(mp, new String[] { args[0], m });
                        result.append(m);
                        if ((j < r.end) && ((j+1) < args.length)) 
                            result.append(sep);
                        j++;
                    }// while
                }// else
                i = r.endpos + 1;
            }

            else if (c == '*') { // comma-separated list of raw arguments
                for (int j = 1; j < args.length; j++) {
                    result.append(args[j]);
                    if ((j + 1) < args.length) result.append(", ");
                }// while
            }

            else { // arg specified by single digit
                int n = Character.getNumericValue(c);
                if ((n >= 0) && (n < args.length)) {
                    result.append(args[n]);
                } // if
                else if (n == -1) {
                    mp.warning("Defined macro: Invalid sequence: $" + c);
                    result.append('$').append(c);
                }
                else {
                    // given number out of range
                    mp.notice("Missing argument in macro " + args[0]);
                }
            } // else

        } // while

        return result.toString();
    }

    /**
     * Returns expansion string.
     * 
     * @return expansion string
     */
    @Override
    public String toString()
    {
        return expansion;
    }

}
