package pmp;

import java.util.Iterator;
import java.util.LinkedList;

import pmp.io.StringInputSource;

/**
 * Internal structure of macroprocessor which provides input source for parser.
 * <p>
 * Title: PMP: Macroprocessor
 * </p>
 * <p>
 * Description: Java macroprocessor
 * </p>
 * <p>
 * Copyright: Copyright (c) 2005
 * </p>
 * 
 * @author Lud\u011Bk Hlav\u00E1\u010Dek
 * @version 1.0
 */

public final class InputQueue extends Macroprocessor.Component {

    public InputQueue(Macroprocessor mp)
    {
        mp.super();
    }

    /**
     * List of InputSource objects.
     */
    private LinkedList<InputSource> iq = new LinkedList<InputSource>();

    /**
     * Buffer for characters which will be read before retriving new data from
     * {@link #iq} list.
     */
    private StringBuilder inputBuffer = new StringBuilder(8096);

    /** Number of bytes processed in this session. */
    private int committedBytes = 0;

    /**
     * Buffer for chars that have been read by {@link #getChar()} but not yet
     * commited.
     */
    private StringBuilder prefetchBuffer = new StringBuilder(64);

    /**
     * Adds new source in front of queue.
     * <p>
     * Note: Prefetch queue must be empty.
     * </p>
     * 
     * @param data
     *            source object
     */
    public InputSource pushFront(InputSource data)
    {
        assert (prefetchBuffer.length() == 0);

        if (inputBuffer.length() > 0) {
            iq.addFirst(new StringInputSource(inputBuffer));
            inputBuffer.setLength(0);
        } // if
        if (data != null) {
            iq.addFirst(data);
        }
        return data;
    }

    /**
     * Adds new source in front of queue. Given {@link CharSequence} is
     * encapsulated by StringInputSource class and
     * {@link #pushFront(InputSource)} is called.
     * <p>
     * Note: Prefetch must be empty.
     * </p>
     * 
     * @param data
     *            source object
     * @return newly created object implementing InputSource interface
     */
    public InputSource pushFront(CharSequence data)
    {
        if ((data != null) && (data.length() > 0)) { return pushFront(new StringInputSource(
                data)); }
        return null;
    }

    public InputSource pushBack(InputSource data)
    {
        if (data != null) {
            iq.addLast(data);
        }
        return data;
    }

    public InputSource pushBack(CharSequence data)
    {
        if ((data != null) && (data.length() > 0)) { return pushBack(new StringInputSource(
                data)); }
        return null;
    }

    /**
     * Reads from input source from input queue and places result into
     * {@link #inputBuffer}.
     * 
     * @return <code>true</code> if at least one char was moved into
     *         {@link #inputBuffer}
     */
    private boolean readInputToBuffer()
    {
        while ((!iq.isEmpty()) && (inputBuffer.length() == 0)) {

            InputSource idata = iq.getFirst();
            String srcdata = idata.read(128);
            if (srcdata.equals("")) {
                (iq.removeFirst()).close();
            }
            else {
                inputBuffer.append(srcdata);
            } // if/else
        } // while
        return (inputBuffer.length() > 0);
    }

    /**
     * Reads single character from input.
     * 
     * @return next character from input, or -1 if end of stream encountered
     */
    public int getChar()
    {
        if (inputBuffer.length() == 0) {
            if (!readInputToBuffer()) { return -1;
            /** do not throw exception */
            }
        }
        char c = inputBuffer.charAt(0);
        inputBuffer.delete(0, 1);
        prefetchBuffer.append(c);
        return c;
    }

    /**
     * Throws away all characters processed by {@link #getChar()}.
     */
    void commit()
    {
        committedBytes += prefetchBuffer.length();
        prefetchBuffer.setLength(0);
    }

    /**
     * Throws away first <code>length</code> characters processed by
     * {@link #getChar()}. Remaining characters are returned to InputQueue.
     * 
     * @param length
     *            number of characters to be committed
     */
    void commit(int length)
    {
        assert (prefetchBuffer.length() >= length);

        committedBytes += length;
        inputBuffer.insert(0, prefetchBuffer.substring(length));
        prefetchBuffer.setLength(0);
    }

    int getCommittedBytes()
    {
        return committedBytes;
    }

    /**
     * Returns true when input queue is empty. While there is a InputSource
     * object in queue, it tries to read some data from it. This method might
     * block (eg: if input is connected to console).
     * 
     * @return <code>true</code> if there is not any character available
     */
    public boolean isEmpty()
    {
        if (inputBuffer.length() > 0) { return true; } // if
        return (!readInputToBuffer());
    }

    /**
     * Clears input queue and closes InputSource objects and all input buffers
     * 
     * @see InputSource
     * 
     */
    public void reset()
    {
        Iterator<InputSource> it = iq.iterator();
        while (it.hasNext()) {
            (it.next()).close();
        } // for

        iq.clear();
        committedBytes = 0;
        prefetchBuffer.setLength(0);
        inputBuffer.setLength(0);
    }

}
