package pmp;

import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Maintains list of macros.
 * <p>
 * Title: PMP: Macroprocessor
 * </p>
 * <p>
 * Description: Java macroprocessor
 * </p>
 * <p>
 * Copyright: Copyright (c) 2005
 * </p>
 * 
 * @author Lud\u011Bk Hlav\u00E1\u010Dek
 * @version 1.0
 */

public class MacroList extends Macroprocessor.Component {

   public MacroList(Macroprocessor mp)
   {
      mp.super();
      modifyListeners = new ArrayList<MacroListModifyListener>(4);
   }

   /** list of macros */
   protected Map<String, MacroListItem> macros = new Hashtable<String, MacroListItem>(
         256); // Hashtable of MacroListItem

   /** list of listeners for macro changes */
   protected List<MacroListModifyListener> modifyListeners;

   /**
    * Adds new listener for macro list modification events.
    * 
    * @param listener
    *           new listener
    */
   public void addModifyListener(MacroListModifyListener listener)
   {
      if (modifyListeners.contains(listener)) {
         //notice("Adding duplicate listener: " + listener.toString());
      } else {
         modifyListeners.add(listener);
      }
   }

   /**
    * Removes all listeners.
    * 
    */
   public void removeModifyListeners()
   {
       modifyListeners.clear();
   }

   /**
    * Sends notification message to all listeners.
    * 
    * @param macroName
    *           name of macro that was modified
    */
   private void notifyListenes(String macroName)
   {
      MacroListModifyEvent event = new MacroListModifyEvent(this, macroName);
      Iterator<MacroListModifyListener> it = modifyListeners.listIterator();
      while (it.hasNext()) {
         (it.next()).macroListModified(event);
      } // while
   }

   /**
    * Returns macro object assicated with given name.
    * 
    * @param macroName
    *           name of macro
    * @return macro object
    */
   public AbstractMacro getMacro(String macroName)
   {
      MacroListItem it = macros.get(macroName);
      if (it != null) {
         return it.getDefinition();
      }
      return null;
   }

   /**
    * Returns macro object assicated with given name. This method performs
    * case-insensitive search, which takes much more, so it should be used only
    * when absolutely necessary.
    * 
    * @param macroName
    *           name of macro
    * @return macro object
    */
   public AbstractMacro getMacroCI(String macroName)
   {
      if (macroName != null) {

         if (macroName.toLowerCase().equals(macroName.toUpperCase())) {
            return getMacro(macroName); // upper- and lower- case strings are
            // equal
         } // if // so case-insensitive search is not required

         Iterator<String> it = macros.keySet().iterator();
         while (it.hasNext()) {
            String key = it.next();
            if (macroName.equalsIgnoreCase(key)) {
               return (macros.get(key)).getDefinition();
            } // if
         } // while
      } // if
      return null;
   }

   /**
    * Removes all macro definitions.
    * 
    * @param all
    *           if <code>true</code> read-only macros will be removed as well
    */
   public void clear(boolean all)
   {
      if (all)
         macros.clear();
      else {
         Set<String> mlist = macros.keySet();
         Iterator<String> mi = mlist.iterator();
         while (mi.hasNext()) {
            MacroListItem mdef = macros.get(mi.next());
            if (!mdef.isReadOnly()) {
               mi.remove();
            }
         }
      }// else
   }

   /**
    * Removes macro definition by calling
    * {@link #setMacro(String,AbstractMacro)} for given <code>macroName</code>
    * with <code>null</code> as macro definition.
    * 
    * @param macroName
    *           name of macro to be removed
    * @return last definition of macro
    */
   public AbstractMacro unsetMacro(String macroName)
   {
      return setMacro(macroName, (AbstractMacro) null);
   }

   /**
    * Copies macro. New macro is identical and independent from the original
    * one.
    * <p>
    * Prints warning on failure (read-only macro or missing source macro).
    * </p>
    * 
    * @param srcName
    *           source macro
    * @param destName
    *           destination name
    * @return <code>true</code> if everything is ok
    */
   public boolean copyMacro(String srcName, String destName)
   {
      MacroListItem srcItem = macros.get(srcName);
      MacroListItem destItem = macros.get(destName);
      if ((destItem != null) && destItem.isReadOnly()) {
         warning("Trying to redefine (using copyMacro) read-only macro "
               + destName);
         return false;
      } // if
      if (srcItem == null) {
         warning("Trying to copy undefined macro " + srcName);
         return false;
      }
      macros.put(destName, srcItem.duplicate());

      notifyListenes(destName);

      return true;
   }

   /**
    * Adds new user-defined macro. Replaces old definition with new one and
    * clears definition stack. This method creates new {@link DefinedMacro}
    * class and calls {@link #setMacro(String,AbstractMacro)}
    * 
    * @see #setMacro(String,AbstractMacro)
    * @param macroName
    *           name of macro
    * @param macro
    *           definition of macro
    * @return returns previous definition
    */
   public AbstractMacro setMacro(String macroName, String macro)
   {
      if (macro != null) {
         AbstractMacro def = new DefinedMacro(macro);
         return setMacro(macroName, def);
      }
      throw new InternalMPError(0x20122);
   }

   /**
    * Defines macro. Replaces old definition with new one and clears definition
    * stack.
    * 
    * @param macroName
    *           name of macro
    * @param macro
    *           definition of macro
    * @return returns previous definition
    */
   public AbstractMacro setMacro(String macroName, AbstractMacro macro)
   {
      MacroListItem item = macros.get(macroName);
      if ((item != null) && item.isReadOnly()) {
         warning("Trying to redefine (using define) read-only macro "
               + macroName);
         return null;
      } // if
      if (macro == null) { // adding null macro would raise NullPointerException
         macros.remove(macroName); // so we just remove previous definition
      } else {
         macros.put(macroName, new MacroListItem(macro, null));
      } // if/else

      notifyListenes(macroName);

      if (item != null) {
         return item.getDefinition();
      }
      return null;
   }

   /**
    * Add all macros from given Map to this MacroList.
    * 
    * @param m
    *           <code>Map</code> with macro definitions
    * @return always <code>true</code>
    */
   public boolean addAll(Map<String, AbstractMacro> m)
   {
      Iterator<Map.Entry<String, AbstractMacro>> it = m.entrySet().iterator();
      while (it.hasNext()) {
         Map.Entry<String, AbstractMacro> e = it.next();
         setMacro(e.getKey(), e.getValue());
      } // while
      return true;
   }

   /**
    * Returns map with all defined macros.
    * 
    * @return Map with macros
    */
   public Map<String, AbstractMacro> getAll()
   {
      Map<String, AbstractMacro> result = new HashMap<String, AbstractMacro>(
            2 * macros.size());
      Set<String> mlist = macros.keySet();
      for (String macro : mlist) {
         result.put(macro, macros.get(macro).getDefinition());
      }// for
      return result;
   }

   /**
    * Replaces current macro definition with new one and stores previous one to
    * definition stack.
    * 
    * @param macroName
    *           name of macro
    * @param macro
    *           new definition
    */
   public void pushMacro(String macroName, AbstractMacro macro)
   {
      if (macros.containsKey(macroName)) {
         MacroListItem item = macros.get(macroName);
         if (item.isReadOnly()) {
            warning("Trying to redefine (using pushdef) read-only macro "
                  + macroName);
            return;
         } // if
         item.pushDef(macro);
      } // if
      else {
         macros.put(macroName, new MacroListItem(macro, false));
      }
      notifyListenes(macroName);
   }

   /**
    * Restores previous macro definition and returns previous one.
    * 
    * @param macroName
    *           name of macro
    * @return previous macro definition
    */
   public AbstractMacro popMacro(String macroName)
   {
      MacroListItem item = macros.get(macroName);
      if (item == null) {
         notice("Trying to redefine (using popdef) non-existing macro "
               + macroName);
         return null;
      } // if
      if (item.isReadOnly()) {
         warning("Trying to redefine (using popdef) read-only macro "
               + macroName);
         return null;
      } // if
      AbstractMacro result = item.popDef();
      if (item.getDefinition() == null) {
         macros.remove(macroName);
      } // if
      notifyListenes(macroName);
      return result;
   }

   /**
    * Returns <code>true</code> when specified directive is defined.
    * 
    * @param directive
    *           name of directive (without prefix)
    * @return <code>true</code> if is defined
    */
   public boolean isDirectiveDefined(String directive)
   {
      return isDefined(forName(directive));
   }

   /**
    * Returns <code>true</code> when specified directive is defined.
    * 
    * @param template
    *           template for name of directive (without prefix)
    * @param id
    *           identifier of directive
    * @return <code>true</code> if is defined
    */
   public boolean isDirectiveDefined(String template, String id)
   {
      return isDefined(forName(template, id));
   }

   /**
    * Returns <code>true</code> if macro with specified name exists.
    * 
    * @param macroName
    *           nema of macro
    * @return <code>true</code> if macro exists
    */
   public boolean isDefined(String macroName)
   {
      return macros.containsKey(macroName);
   }

   /**
    * Returns <code>true</code> if macro is marked as read-only. If macro
    * doesn't exists, <code>false</code> is returned.
    * 
    * @param macroName
    *           name of macro
    * @return <code>true</code> if macro exists and is marked as read-only.
    */
   public boolean isReadOnly(String macroName)
   {
      MacroListItem it = macros.get(macroName);
      if (it != null) {
         return it.isReadOnly();
      }
      return false;
   }

   /**
    * Returns macro as string value. This method DOES NOT prepends prefix.
    * 
    * @param macroName
    *           name of macro
    * @return String representation of macro
    */
   private String getMacroAsString(String macroName)
   {
      AbstractMacro macro = getMacro(macroName);
      if (macro != null) {
         return macro.toString();
      }
      return "";
   }

   /**
    * Returns macro as integer value. This method DOES NOT prepends prefix.
    * 
    * @param macroName
    *           name of macro
    * @return integer representation of macro
    */
   private int getMacroAsInt(String macroName)
   {
      String strValue = "0";
      AbstractMacro macro = getMacro(macroName);
      if (macro != null) {
         strValue = macro.toString();
      } // if
      int result;
      try {
         result = Integer.parseInt(strValue);
      } catch (NumberFormatException ex) {
         result = 0; // return zero when content is not number
      }
      return result;
   }

   /**
    * Returns macro as boolean value. This method DOES NOT prepends prefix.
    * 
    * @param macroName
    *           name of macro
    * @return boolean representation of macro
    */
   private boolean getMacroAsBool(String macroName)
   {
      AbstractMacro macro = getMacro(macroName);
      if (macro == null)
         return false;

      final String val = macro.toString().toLowerCase();
      if ("".equals(val) || "0".equals(val) || "off".equals(val)
            || "no".equals(val) || "false".equals(val)) {
         return false;
      }
      return true;
   }

   /**
    * Returns name of configuration directive for given name.
    * 
    * @param var
    *            name of directive name (typicaly obained from static field of
    *            {@link ConfigDirectives} class)
    * @return full directive name (including prefix)
    */
   public String forName(String var)
   {
       return getMP().getPrefix() + var;
   }

   /**
    * Returns name of configuration directive for given template and id.
    * 
    * @param var
    *            template for directive name (typically obained from static
    *            field of {@link ConfigDirectives} class)
    * @param id
    *            identifier
    * @return full directive name (including prefix) after replacing "*" with
    *         id
    */
   public String forName(String var, String id)
   {
       int index = var.indexOf('*');
       assert (index != -1);

       StringBuilder result = new StringBuilder(64);
       result.append(getMP().getPrefix()).append(var.substring(0, index))
               .append(id).append(var.substring(index + 1));

       return result.toString();
   }

   /**
    * Defines macro. Its name is created from <code>directive</code>
    * with prependef prefix.

    * @see #setMacro(String, String)
    * @see #forName(String)
    * 
    * @param directive Name of configuration directive
    * @param macro macro object 
    * @return previous macro definition as returned by {@link #setMacro(String, AbstractMacro)}
    */
   public AbstractMacro setMacroForName(String directive, AbstractMacro macro)
   {
       String name = getMP().getPrefix() + directive;
       return setMacro(name, macro);
   }

   /**
    * Defines macro. Its name is created from <code>directive</code>
    * with prependef prefix.

    * @see #setMacro(String, String)
    * @see #forName(String)
    * 
    * @param directive Name of configuration directive
    * @param macro macro definition 
    * @return previous macro definition as returned by {@link #setMacro(String, String)}
    */
   public AbstractMacro setMacroForName(String directive, String macro)
   {
       String name = getMP().getPrefix() + directive;
       return setMacro(name, macro);
   }

   /**
    * Returns configuration directive as string value. This method prepends
    * prefix using {@link #forName(String)}.
    * <p>
    * Return value is obtained by calling macro's toString method
    * </p>
    * <p>
    * Returns <code>""</code> when given directive is not defined.
    * </p>
    * 
    * @param directive
    *           name of configuration directive
    * @return value of configuration directive
    * @see pmp.AbstractMacro#toString()
    */
   public String getString(String directive)
   {
      return getMacroAsString(forName(directive));
   }

   /**
    * Returns configuration directive as string value. This method prepends
    * prefix using {@link #forName(String,String)}.
    * <p>
    * Returns <code>""</code> when given directive is not defined.
    * </p>
    * @param directive_template
    *           template for name of configuration directive
    * @param id
    *           identifier part of configuration directive
    * @return value of configuration directive
    */
   public String getString(String directive_template, String id)
   {
      return getMacroAsString(forName(directive_template, id));
   }

   /**
    * Returns result of {@link #getString(String)} after being converted to
    * lower-case.
    * 
    * @param directive
    *           name of configuration directive
    * @return lower-cased value of configuration directive
    */
   public String getStringL(String directive)
   {
      return getString(directive).toLowerCase();
   }

   /**
    * Returns result of {@link #getString(String)} after being converted to
    * lower-case.
    * 
    * @param directive_template
    *           temlate for name of configuration directive
    * @param id
    *           identifier of configuration directive
    * @return lower-cased value of configuration directive
    */
   public String getStringL(String directive_template, String id)
   {
      return getString(directive_template, id).toLowerCase();
   }

   /**
    * Returns boolean value associated with given directive name. Any non-empty
    * string is considered as <code>true</code> except strings: "off",
    * "false", "no", "0" (case-insensitively)
    * 
    * @param directive
    *           name of configuration directive
    * @return boolean value of given directive
    */
   public boolean getBool(String directive)
   {
      return getMacroAsBool(forName(directive));
   }

   /**
    * Returns boolean value associated with given directive name. Any non-empty
    * string is considered as <code>true</code> except strings: "off",
    * "false", "no", "0" (case-insensitively)
    * 
    * @param template
    *           tempate for name of configuration directive
    * @param id
    *           identifier of directive
    * @see #forName(String,String)
    * @return boolean value of given directive
    */
   public boolean getBool(String template, String id)
   {
      return getMacroAsBool(forName(template, id));
   }

   /**
    * Returns integer value associated with given directive name. If macro
    * doesn't conatain integer value, zero is returned.
    * 
    * @param directive
    *           name of configuration directive
    * @return integer value of given directive
    */
   public int getInt(String directive)
   {
      return getMacroAsInt(forName(directive));
   }

   /**
    * Returns integer value associated with given directive name. If macro
    * doesn't conatain integer value, zero is returned.
    * 
    * @param template
    *           tempate for name of configuration directive
    * @param id
    *           identifier of directive
    * @return integer value of given directive
    */
   public int getInt(String template, String id)
   {
      return getMacroAsInt(forName(template, id));
   }

   /**
    * Returns array of indices for given mask. PMP prefix is automaticaly
    * prepended.
    * <p>
    * Note: Identifier containing character '.' (&#x5c;u002e) will not be found.
    * </p>
    * 
    * @param mask
    *           mask (without prefix)
    * @return array of indices
    */
   public String[] getIDsByMask(String mask)
   {
      List<String> ids = new ArrayList<String>(32);
      int apos = mask.indexOf('*');

      assert (apos >= -1); // this should never happen

      String regexp = "\\Q" + getMP().getPrefix() + mask.substring(0, apos)
            + "\\E([^.]*)\\Q" + mask.substring(apos + 1) + "\\E";
      Pattern p = RegExps.compilePattern(regexp);
      Matcher m = p.matcher("");

      Iterator<String> it = macros.keySet().iterator();
      while (it.hasNext()) {
         String key = it.next();
         m.reset(key);
         if (m.matches()) {
            ids.add(m.group(1)); // extract index and add it to list
         } // id
      } // while

      String[] result = new String[ids.size()];
      ids.toArray(result);
      Arrays.sort(result);
      return result;

   }

   /**
    * Returns array of macro names.
    * 
    * @return iterator object
    */
   public String[] getMacroNames()
   {
      return macros.keySet().toArray(new String[0]);
   }

   /**
    * Returns read-only iterator over macro names.
    * 
    * @return iterator object
    */
   public Iterator<String> macroIterator()
   {
      return macros.keySet().iterator();
   }

}
