package pmp;

import java.util.*;

/**
 * Element of list of macros. Encapsules definition of macro, definition stack,
 * and attributes.
 * <p>
 * Title: PMP: Macroprocessor
 * </p>
 * <p>
 * Description: Java macroprocessor
 * </p>
 * <p>
 * Copyright: Copyright (c) 2005
 * </p>
 * 
 * @author Lud\u011Bk Hlav\u00E1\u010Dek
 * @version 1.0
 * @todo make use of executable flag
 */

public class MacroListItem {

   private AbstractMacro def;

   private final LinkedList<AbstractMacro> stack;

   private final boolean readOnly;

   private final boolean executable;

   /**
    * Returns new instance of MacroListItem with identical contents as this
    * object.
    * 
    * @return copy of this object
    */
   public MacroListItem duplicate()
   {
      return new MacroListItem(this.def, this.stack, this.readOnly,
            this.executable);
   }

   /**
    * Conveninece constructor.
    * 
    * @param def
    *           macro definition object
    * @param stack
    *           macro definition history
    */
   public MacroListItem(AbstractMacro def, Collection<AbstractMacro> stack)
   {
      this(def, stack, false, true);
   }

   /**
    * Convenience constructor
    * 
    * @param def
    *           macro definition
    * @param readOnly
    *           read-only flag
    */
   public MacroListItem(AbstractMacro def, boolean readOnly)
   {
      this(def, null, readOnly, true);
   }

   /**
    * Main constructor.
    * 
    * @param def
    *           macro definition object
    * @param stack
    *           definition history
    * @param readOnly
    *           read-only flag
    */
   public MacroListItem(AbstractMacro def, Collection<AbstractMacro> stack,
         boolean readOnly, boolean executable)
   {
      this.def = def;
      this.readOnly = readOnly;
      this.executable = executable;
      this.stack = new LinkedList<AbstractMacro>();
      if (stack != null) {
         this.stack.addAll(stack);
      }
   }

   /**
    * Returns current macro definition.
    * 
    * @return macro object
    */
   public AbstractMacro getDefinition()
   {
      return def;
   }

   public boolean getExecutableFlag()
   {
      return this.executable;
   }

   public boolean getReadOnlyFlag()
   {
      return this.readOnly;
   }

   /**
    * Replaces current macro definition and saves current definition to stack.
    * 
    * @param macro
    *           new definition
    */
   public void pushDef(AbstractMacro macro)
   {
      assert (stack != null);
      stack.addFirst(def);
      def = macro;
   }

   public AbstractMacro popDef()
   {
      AbstractMacro result = def;
      if (stack.isEmpty()) {
         def = null;
      } else {
         def = stack.removeFirst();
      }
      return result;
   }

   /**
    * Returns <code>true</code> if this definition is immutable.
    * 
    * @return state of read-only flag
    */
   public boolean isReadOnly()
   {
      return readOnly;
   }

}
