package pmp;

import java.util.EventObject;

/**
 * Event class for {@link MacroListModifyListener}.
 * <p>
 * Title: PMP: Macroprocessor
 * </p>
 * <p>
 * Description: Java macroprocessor
 * </p>
 * <p>
 * Copyright: Copyright (c) 2005
 * </p>
 * 
 * @author Luděk Hlaváček
 * @version 1.0
 */

public class MacroListModifyEvent extends EventObject {

   /** Mask for everything but "pmp:…" */
   public static final int MASK_NON_PMP = 0x020000;

   /** Mask for "pmp://…" */
   public static final int MASK_PMP_SCHEME = 0x010000;

   /** Mask for "pmp:…" */
   public static final int MASK_PMP = 0x0001;

   /** Mask for "pmp:prefix" */
   public static final int MASK_PMP_PREFIX = 0x040000 | MASK_PMP;

   /** Mask for "pmp:string&#46;…" */
   public static final int MASK_PMP_STRING = 0x0002 | MASK_PMP;

   // /** Mask for "pmp:symbols&#46;…" */
   // public static final int MASK_PMP_SYMBOLS = 0x0004 | MASK_PMP;

   /** Mask for "pmp:macro&#46;…" */
   public static final int MASK_PMP_MACRO = 0x0008 | MASK_PMP;

   /** Mask for "pmp:code&#46;…" */
   public static final int MASK_PMP_CODE = 0x0010 | MASK_PMP;

   /** Mask for "pmp:stack&#46;…" */
   public static final int MASK_PMP_STACK = 0x0020 | MASK_PMP;

   /** Mask for "pmp:last&#46;…" */
   public static final int MASK_PMP_LOG_EVENT = 0x0040 | MASK_PMP;

   /** Mask for "pmp:input&#46;…" */
   public static final int MASK_PMP_INPUT = 0x0080 | MASK_PMP;

   /** Mask for "pmp:output&#46;…" */
   public static final int MASK_PMP_OUTPUT = 0x0100 | MASK_PMP;

   /** <code>{@link #MASK_PMP_INPUT} | {@link #MASK_PMP_OUTPUT}</code> */
   public static final int MASK_PMP_IO = MASK_PMP_INPUT | MASK_PMP_OUTPUT;

   /**
    * Flags for this event. Calculated on demand by updateMask.
    * 
    * @see #updateMask
    */
   private int mask;

   /**
    * Name of macro which was modified.
    */
   private final String macroName;

   /**
    * Creates event object for given name and source
    * 
    * @param source
    *           source MacroList
    * @param macroName
    *           name of macro which invoked this event
    */
   public MacroListModifyEvent(MacroList source, String macroName)
   {
      super(source);
      this.mask = -1;
      this.macroName = macroName;
   }

   /**
    * If this event was invoked by changing value of a configuration directive
    * this method returns its name (without prefix). Otherwise returns empty
    * string. Returns name of directive
    * 
    * @return name of directive or empty string
    */
   public String getDirectiveName()
   {
      String prefix = ((Macroprocessor.Component) source).getMP().getPrefix();
      if (macroName.startsWith(prefix)) {
         return macroName.substring(prefix.length());
      }
      return "";
   }

   private void updateMask()
   {
      Macroprocessor mp = ((Macroprocessor.Component) source).getMP();

      mask = MASK_NON_PMP;
      String prefix = mp.getPrefix();
      if (macroName.startsWith(prefix)) {
         mask = MASK_PMP;
         String section = macroName.substring(prefix.length());
         if (section.startsWith("string.")) {
            mask |= MASK_PMP_STRING;
         } else if (section.startsWith("stack.")) {
            mask |= MASK_PMP_STACK;
         } else if (section.startsWith("input.")) {
            mask |= MASK_PMP_INPUT;
         }
         /*
          * else if (section.startsWith("symbols.")) { mask |= MASK_PMP_SYMBOLS; }
          */
         else if (section.startsWith("output.")) {
            mask |= MASK_PMP_OUTPUT;
         } else if (section.startsWith("code.")) {
            mask |= MASK_PMP_CODE;
         } else if (section.startsWith("macro.")) {
            mask |= MASK_PMP_MACRO;
         } else if (section.startsWith("last.")) {
            mask |= MASK_PMP_LOG_EVENT;
         } else if (section.equals(ConfigDirectives.PMP_PREFIX)) {
            mask |= MASK_PMP_PREFIX;
         }
      } // if
      if (macroName.startsWith("pmp://")) {
         mask = MASK_PMP_SCHEME;
      }
   }

   public int getMask()
   {
      if (mask == -1) {
         updateMask();
      } // if
      return mask;
   }

   /**
    * Tests if all of given bits are set in mask
    * 
    * @param maskFlag
    *           bits
    * @return <code>true</code> if all bits are set.
    */
   public boolean isFlagSet(int maskFlag)
   {
      if (mask == -1) {
         updateMask();
      } // if
      return ((mask & maskFlag) == maskFlag);
   }

   /**
    * Returns a String representation of this EventObject.
    * 
    * @return A String representation of this EventObject.
    */
   @Override
   public String toString()
   {
      return getClass().getName() + "[macro=" + macroName + "; mask=" + mask
            + "]";
   }

   private static final long serialVersionUID = -1L;

}
