package pmp;

import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Locale;

import static pmp.ConfigDirectives.*;

/**
 * Main class of Macroprocessor.
 * <p>
 * Title: PMP: Macroprocessor
 * </p>
 * <p>
 * Description: Java macroprocessor
 * </p>
 * <p>
 * Copyright: Copyright (c) 2005
 * </p>
 * 
 * @author Lud\u011Bk Hlav\u00E1\u010Dek
 * @version 1.0
 */

public class Macroprocessor implements MacroListModifyListener {

    public Macroprocessor()
    {
        macros = new MacroList(this);
        mpiq = new InputQueue(this);
        mpan = new Analyzer(this);
        mpio = new InputOutput(this); // dummy IO
    }

    public void macroListModified(MacroListModifyEvent event)
    {
        if (event.isFlagSet(MacroListModifyEvent.MASK_PMP_PREFIX)) {
            String newprefix = macros.getString(PMP_PREFIX);
            pmpprefix = newprefix;
        }
    }

    /**
     * Assigns InputOutput class to macroprocessor.
     * <p>
     * <em> This should be done as soon as possible since it is the only way MP
     * can provide any feedback.</em>
     * </p>
     * 
     * @param io
     *            InputOutput class
     */
    public void assignIO(InputOutput io)
    {
        this.mpio = io;
        InputSource def = mpio.getDefaultInput();
        if (def != null) mpiq.pushBack(def);
    }

    /**
     * Returns <code>true</code> if macroprocessor is ready. This method
     * doesn't check whether InputQueue is not empty.
     * 
     * @return true if it is safe to call {@link #parse()}.
     */
    public boolean isReady()
    {
        boolean test = true;
        test &= (mpio != null);
        return test;
    }

    /**
     * Resets macroprocessor.
     */
    public void reset()
    {
        macros.removeModifyListeners();
        macros.clear(true);
        setPrefix(DEFAULT_PREFIX);
        macros.addModifyListener(this);
        mpiq.reset();
    }

    /**
     * Starts macroprocessor.
     * 
     * @throws FatalParserException
     *             when something fails and it is impossible to continue
     * @throws InternalMPError
     *             when internal error has occured
     * @throws FatalParserException
     *             when parsing error has occurs
     */
    public void parse() throws FatalParserException, InternalMPError
    {
        if (isReady()) {
            mpan.processInput();
        }
        else {
            throw new FatalParserException("Macroprocessor is not ready.");
        }
    }

    protected MacroList macros;

    protected InputOutput mpio;

    protected InputQueue mpiq;

    protected Analyzer mpan;

    public final Macroprocessor getMP()
    {
        return Macroprocessor.this;
    }

    public final InputOutput getMPIO()
    {
        return mpio;
    }

    public final InputQueue getMPIQ()
    {
        return mpiq;
    }

    public final Analyzer getMPAnalyzer()
    {
        return mpan;
    }

    public final MacroList getMPMacros()
    {
        return macros;
    }

    public static final String DEFAULT_PREFIX = "pmp:";

    protected String pmpprefix = DEFAULT_PREFIX;

    /**
     * Returns prefix of configuration directives.
     * 
     * @see pmp.ConfigDirectives
     * @return current prefix of configuration directives
     */
    public final String getPrefix()
    {
        return pmpprefix;
    }

    /**
     * Changes prefix of configuration directives.
     * 
     * @see pmp.ConfigDirectives
     * @param newprefix
     *            new prefix for configuration directives
     * @return previous prefix
     */
    public final String setPrefix(String newprefix)
    {
        String result = pmpprefix;
        // pmpprefix = newprefix; -- this is done by listener
        macros.setMacro(newprefix + PMP_PREFIX, newprefix);
        return result;
    }

    /** Level of debugging output. */
    private int debugLevel = DEBUG_NODEBUG;

    /** Debug level: No debugging */
    public static final int DEBUG_NODEBUG = 0;

    /** Debug level: Print notice messages */
    public static final int DEBUG_NOTICES = 5;

    /** Debug level: Print macro calls */
    public static final int DEBUG_MACRO_CALLS = 10;

    /** Debug level: Print macro call results */
    public static final int DEBUG_MACRO_RESULTS = 15;

    /** Debug level: Print every processed token */
    public static final int DEBUG_EVERY_TOKEN = 20;

    /**
     * Get current level of debugging output.
     * 
     * @return debug level
     */
    public final int getDebugLevel()
    {
        return debugLevel;
    }

    /**
     * Sets new debugging level.
     * 
     * @param level
     *            level of debugging
     */
    public final void setDebugLevel(int level)
    {
        this.debugLevel = level;
        macros.setMacroForName(DEBUGLEVEL, Integer
                .toString(debugLevel));
    }

    /**
     * Sets debugging level according to given string. This mathod accepts names
     * of constants and decimal integer values.
     * 
     * @param level
     *            desired level as String
     * @return <code>true</code> if <code>level</code> parameter contains
     *         valid value
     */
    public final boolean setDebugLevel(String level)
    {
        if ("EVERY_TOKEN".equalsIgnoreCase(level)) {
            debugLevel = DEBUG_EVERY_TOKEN;
        }
        else if ("MACRO_CALLS".equalsIgnoreCase(level)) {
            debugLevel = DEBUG_MACRO_CALLS;
        }
        else if ("MACRO_RESULTS".equalsIgnoreCase(level)) {
            debugLevel = DEBUG_MACRO_RESULTS;
        }
        else if ("DEBUG_NOTICES".equalsIgnoreCase(level)) {
            debugLevel = DEBUG_NOTICES;
        }
        else if ("DEBUG_NODEBUG".equalsIgnoreCase(level)) {
            debugLevel = DEBUG_NODEBUG;
        }
        else {
            try {
                this.debugLevel = Integer.parseInt(level);
            } catch (NumberFormatException ex) {
                return false;
            }
        }
        macros.setMacroForName(DEBUGLEVEL, Integer
                .toString(debugLevel));
        return true;
    }

    public void debug(Object message, int level)
    {
        if (this.debugLevel >= level) {
            mpio.debugHandler(message.toString());
        } // if
    }

    protected Locale currentLoc = Locale.getDefault();

    protected String currentLocStr = currentLoc.toString();

    public Locale getLocale()
    {
        String loc = macros.getString(LOCALE);
        if (loc.equals(currentLoc)) return currentLoc;

        currentLocStr = loc;
        currentLoc = Locale.getDefault();
        if (!"".equals(loc)) {
            String[] locs0 = loc.split("_");
            String[] locs = new String[3 + locs0.length]; // make sure locs is
            // at
            // least 3 items long
            Arrays.fill(locs, "");
            System.arraycopy(locs0, 0, locs, 0, locs.length);
            currentLoc = new Locale(locs[0], locs[1], locs[2]);
        }
        Locale.setDefault(currentLoc);
        return currentLoc;
    }

    /**
     * Invokes macro's {@link AbstractMacro#run(Macroprocessor,String[])} method
     * and returns result.
     * 
     * @param item
     *            object describing macro-call
     * @return result of macro execution
     */
    public String runMacro(CallStackItem item)
    {
        item.appendRequiredParams();

        for (int i = 0; i < item.params.size(); i++) {
            macros.setMacro(macros.forName(STACK_ITEM__,
                    Integer.toString(i)), item.getParam(i));
        } // for
        String paramCount = macros.forName(STACK_ITEMS_COUNT);
        String openingBracket = macros
                .forName(STACK_OPENING_BRACKET);
        String closingBracket = macros
                .forName(STACK_CLOSING_BRACKET);
        macros.setMacro(paramCount, Integer.toString(item.params.size()));
        if (item.openingBracket != null) {
            macros.setMacro(openingBracket, item.openingBracket);
        }
        if (item.closingBracket != null) {
            macros.setMacro(closingBracket, item.closingBracket);
        }

        String result = item.getMacro().run(this, item.getStrParams());

        macros.unsetMacro(paramCount);
        for (int i = 0; i < item.params.size(); i++) {
            macros.unsetMacro(macros.forName(STACK_ITEM__,
                    Integer.toString(i)));
        } // for
        macros.unsetMacro(openingBracket);
        macros.unsetMacro(closingBracket);

        return result;
    }

    /**
     * Adds given library to library list.
     * 
     * @param name
     *            name of library
     */
    public void importLibrary(String name)
    {
        try {
            if (name.startsWith("http://")) {
                // nothing
            }
            else if (name.startsWith("file:")) {
                // nothing
            }
            else {
                name = "file:" + name;
            }
            debug("Loading library " + name + "...", DEBUG_NOTICES);
            URL url = new URL("jar:" + name + "!/");
            urls.add(url);
        } catch (MalformedURLException e) {
            error("Could not load library " + name);
        }
    }

    /**
     * List of libraries with macro class files.
     */
    protected ArrayList<URL> urls = new ArrayList<URL>(8);

    /**
     * Loads macro class. Given class name must be subclass of
     * {@link AbstractMacro}. If something fails, warning is produced.
     * 
     * @see #warning(String)
     * @param className
     *            name of class
     * @return <code>null</code> if loading fails, otherwise instance of
     *         loaded class.
     */
    public AbstractMacro loadFromClass(String className)
    {
        AbstractMacro macro = null;
        debug("loadFromClass "+className, DEBUG_NOTICES);
        try {
            Class macroClass = null;
            if (urls.size() > 0) {
                URLClassLoader loader1 = 
                    new URLClassLoader(urls.toArray(new URL[0]));
                macroClass = Class.forName(className, true, loader1);
            }
            else {
                macroClass = Class.forName(className);
            }
            Object instance = macroClass.newInstance();
            if (instance instanceof AbstractMacro) {
                macro = (AbstractMacro) instance;
            }
            else {
                warning("Class not subclass of AbstractMacro: " + className);
            }
        } catch (IllegalAccessException ex) {
            warning("Couldn't create instance: " + className);
        } catch (InstantiationException ex) {
            warning("Couldn't create instance: " + className);
        } catch (ClassNotFoundException ex) {
            warning("Couldn't load class: " + className);
        } catch (NoClassDefFoundError err) {
            warning("Couldn't load class: " + className);
        }

        return macro;
    }

    /**
     * Boolean variable that indicates call to error/warning/notice methods. If
     * <code>true</code> these methods must not call code that could
     * recursively invoke one of them.
     */
    private boolean reported = false;

    /**
     * Produces fatal error. Config directive
     * {@link ConfigDirectives#LAST_ERROR} is updated and given message is sent
     * to InputOutput object of this macroprocessor. This method terminates
     * macroprocessor by throwing {@link FatalParserException}.
     * 
     * @see pmp.InputOutput#errorHandler(String)
     * @param msg
     *            message text
     */
    public final void error(String msg)
    {
        if (!reported) {
            reported = true;
            getMPMacros().setMacroForName(LAST_ERROR, msg);
            reported = false;
        }
        mpio.errorHandler(msg);
        throw new FatalParserException(
                "Macroprocesor terminated due to fatal error.");
    }

    /**
     * Produces warning message. Config directive
     * {@link ConfigDirectives#LAST_WARNING} is updated and given message is
     * sent to InputOutput object of this macroprocessor.
     * 
     * @see pmp.InputOutput#warningHandler(String)
     * @param msg
     *            message text
     */

    public final void warning(String msg)
    {
        if (!reported) {
            reported = true;
            macros.setMacroForName(LAST_WARNING, msg);
            reported = false;
        }
        mpio.warningHandler(msg);
    }

    /**
     * Produces notice message. Config directive
     * {@link ConfigDirectives#LAST_NOTICE} is updated and given message is sent
     * to InputOutput object of this macroprocessor.
     * 
     * @see pmp.InputOutput#noticeHandler(String)
     * @param msg
     *            message text
     */
    public final void notice(String msg)
    {
        if (!reported) {
            reported = true;
            getMPMacros().setMacroForName(LAST_NOTICE, msg);
            reported = false;
        }
        mpio.noticeHandler(msg);
    }

    public abstract class Component {

        public final Macroprocessor getMP()
        {
            return Macroprocessor.this;
        }

        public final InputOutput getMPIO()
        {
            return mpio;
        }

        public final InputQueue getMPIQ()
        {
            return mpiq;
        }

        public final Analyzer getMPAnalyzer()
        {
            return mpan;
        }

        public final MacroList getMPMacros()
        {
            return macros;
        }

        public final void notice(String msg)
        {
            Macroprocessor.this.notice(msg);
        }

        public final void warning(String msg)
        {
            Macroprocessor.this.warning(msg);
        }

        public final void error(String msg)
        {
            Macroprocessor.this.error(msg);
        }

    } // Component

}
