package pmp;

import java.util.*;
import static pmp.MatcherIface.*;
import java.util.regex.*;
import java.lang.ref.*;

/**
 * List of regular expressions with associated tokens.
 * <p>
 * Title: PMP: Macroprocessor
 * </p>
 * <p>
 * Description: Java macroprocessor
 * </p>
 * <p>
 * Copyright: Copyright (c) 2005
 * </p>
 * 
 * @author Lud\u011Bk Hlav\u00E1\u010Dek
 * @version 1.0
 */

public class RegExps extends Macroprocessor.Component {

    private static final Map<String, WeakReference<Pattern>> patternCache = new HashMap<String, WeakReference<Pattern>>(
            50);

    private static final Object patternCacheLock = new Object();

    /**
     * Creates Pattern objects using cache if possible.
     * <p>
     * All patterns are compiled with flag
     * {@link java.util.regex.Pattern#DOTALL}.
     * </p>
     * 
     * @param pattern
     *            given regex
     * @return Pattern object for given regex
     * 
     * @see java.util.regex.Pattern#compile(java.lang.String, int)
     */
    public static Pattern compilePattern(String pattern)
    {
        WeakReference<Pattern> ref = null;
        synchronized (patternCacheLock) {
            ref = patternCache.get(pattern);
        }
        Pattern p = null;
        if (ref != null) p = ref.get();
        if (p == null) {
            p = Pattern.compile(pattern, Pattern.DOTALL);
            synchronized (patternCacheLock) {
                patternCache.put(pattern, new WeakReference<Pattern>(p));
            }
        }
        return p;
    }

    private List<MatcherIface> items;

    /**
     * Contains current string.
     */
    private StringBuilder parsedInput = new StringBuilder(128);

    /**
     * Reference to macroprocessor's InputQueue.
     */
    private InputQueue iq;

    public RegExps(Macroprocessor mp)
    {
        mp.super();
        iq = getMPIQ();
        items = new ArrayList<MatcherIface>(32);
    }

    @Override
    public String toString()
    {
        StringBuilder result = new StringBuilder(1024);
        result.append(super.toString()).append(" = {\n");
        for (int i = 0; i < items.size(); i++) {
            MatcherIface re = items.get(i);
            result.append(" ").append(re.toString()).append("\n");
        } // for
        result.append("}\n");
        return result.toString();
    }

    /**
     * Append character to currently analyzed token.
     * 
     * @param c
     *            character
     */
    private final void feed(int c)
    {
        parsedInput.append((char) c);
    }

    private void matchAll()
    {
        Iterator<MatcherIface> it = items.iterator();
        while (it.hasNext()) {
            (it.next()).matches(parsedInput);
        } // while
    }

    /**
     * Returns all unprocessed input and resets all states of Matchers.
     */
    void reset()
    {
        parsedInput.setLength(0);

        Iterator<MatcherIface> it = items.iterator();
        while (it.hasNext()) {
            it.next().reset();
        } // while
    }

    /**
     * Remove all active regular expressions.
     */
    void clear()
    {
        items.clear();
    }

    /**
     * Returns number of active regular expressions.
     * 
     * @return number of inems in internal list of reg. exppressions
     */
    int size()
    {
        return items.size();
    }

    /**
     * Adds new regular expression.
     * 
     * @param regexp
     *            regular expression
     * @param token
     *            Token object that will be returned if input matches
     *            <code>regexp</code>. Field <code>string</code> will
     *            contain matching string.
     */
    void addRegExp(String regexp, Token token)
    {
        if (EmptyMatcher.EMPTY_PATTERN.equals(regexp))
            items.add(new EmptyMatcher(token));
        else if (FailingMatcher.FAILING_PATTERN.equals(regexp))
            items.add(new FailingMatcher(token));
        else {
            items.add(new JavaMatcher(token, compilePattern(regexp)));
        }
    }

    protected static final int CHUNK_SIZE = 16;

    /**
     * Returns next token.
     * 
     * @return token with processed string, type and ID
     */
    public Token outputToken()
    {
        reset();
        while (true) {
            int c = -1;
            int b = 0;
            while (b++ < CHUNK_SIZE) {
                c = iq.getChar();
                if (c == -1) break;
                feed(c);
            } // if

            matchAll();

            State rstate = State.MISMATCH;

            fori: for (int i = 0; i < items.size(); i++) {
                MatcherIface re = items.get(i);
                switch (re.getState()) {

                // expression matches input, but we may consume more chars
                case MATCH:
                    if (c != -1) { // unless its end of file
                        rstate = State.MATCH;
                        break fori;
                    }

                // too much characters read. return valid part
                case OVERFEED:
                    // there must not be R_PARTIAL or R_MATCH with higher
                    // priority
                    if (rstate == State.MISMATCH) {

                        Token t = re.getToken();
                        int len = re.getMatchingLength();
                        t.setString(parsedInput.substring(0, len));
                        iq.commit(len);
                        reset();
                        return t;
                    } // if
                    break;

                // we are in the middle of expression
                case PARTIAL:
                    if (c != -1) {
                        rstate = State.PARTIAL;
                    }
                    break;

                case MISMATCH: // do nothing
                    break;

                default:
                    throw new InternalMPError(0x030154);
                } // switch
            } // for

            // failed to match anything -> return single char
            if (rstate == State.MISMATCH) {
                if (parsedInput.length() > 0) {
                    tokenInvalid.setString(parsedInput.substring(0, 1));
                    iq.commit(1);
                    // reset(); - not req.
                    return tokenInvalid;
                }
                else {
                    // reset(); - not req.
                    return tokenEof;
                }
            } // if

        } // while
    }

    private final Token tokenInvalid = new Token(Token.Type.SINGLE_CHAR);

    private final Token tokenEof = new Token(Token.Type.END);

}
