package pmp;

/**
 * Source code token. It is used as container to transport tokens between
 * RegExps and Analyzer objects. This objects is constructed with constant type
 * (and string type identifier). Field <code>string</code> is variable so that
 * this object can be reused.
 * 
 * <p>
 * Title: PMP: Macroprocessor
 * </p>
 * <p>
 * Description: Java macroprocessor
 * </p>
 * <p>
 * Copyright: Copyright (c) 2005
 * </p>
 * 
 * @author Lud\u011Bk Hlav\u00E1\u010Dek
 * @version 1.0
 */

public class Token {

   public enum Type {
      EMPTY, START, END, SINGLE_CHAR, PLAINTEXT, CODE_START, CODE_END, CODE,
      /**
       * Token is possibly name of macro. Analyzer decides whether it is really
       * name of existing macro.
       */
      MACRO_NAME,
      /** Start of paramlist */
      OPENING_BRACKET,
      /** End of paramlist */
      CLOSING_BRACKET, PARAM_SEPARATOR, STRING_START, STRING_END, STRING_TOKEN, STRING_SPECIAL_SYMBOL
   }// enum Type

   /**
    * String from input text which was recognized as token.
    */
   private String string = "";

   /**
    * Token type
    */
   private final Type type;

   /**
    * Identifier of token type (only for string tokens)
    */
   private final String typeID;

   public Token(Type type)
   {
      this(type, "", "");
   }

   public Token(Type type, String typeID)
   {
      this(type, typeID, "");
   }

   public Token(Type type, String typeID, String string)
   {
      this.type = type;
      if (typeID == null) {
         this.typeID = "";
      } else {
         this.typeID = typeID;
      }
      this.string = string;
   }

   /**
    * Clears token carried by this object.
    */
   public void clearString()
   {
      this.string = "";
   }

   /**
    * Changes token carried by this object.
    * 
    * @param string
    *           new token
    */
   public void setString(String string)
   {
      this.string = string;
   }

   /**
    * Returns token carried by this object.
    * 
    * @return token
    */
   public String getString()
   {
      return this.string;
   }

   /**
    * Returns lexical type of this token.
    * 
    * @return lexical type
    */
   public Type getType()
   {
      return this.type;
   }

   /**
    * Returns type identifier. Valid only for string tokens.
    * 
    * @return type identifier
    */
   public String getTypeID()
   {
      return this.typeID;
   }

   @Override
   public String toString()
   {
      StringBuilder str = new StringBuilder(120);
      str.append("[").append(type.toString());
      if (!"".equals(typeID)) {
         str.append(": ").append(typeID);
      }
      str.append("]");
      if (!"".equals(string)) {
         str.append(" \"").append(string).append("\"");
      } // if
      return str.toString();
   }

}
