/* 
 * Copyright  2006 Luděk Hlaváček
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package pmp.ant;

import java.io.File;
import java.io.IOException;
import java.io.Writer;
import java.util.Enumeration;
import java.util.Hashtable;

import javax.xml.parsers.ParserConfigurationException;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.DirectoryScanner;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.taskdefs.MatchingTask;
import org.apache.tools.ant.types.Mapper;
import org.apache.tools.ant.util.FileNameMapper;
import org.apache.tools.ant.util.FileUtils;
import org.apache.tools.ant.util.IdentityMapper;
import org.apache.tools.ant.util.SourceFileScanner;

import pmp.AbstractMacro;
import pmp.ConfigDirectives;
import pmp.FatalParserException;
import pmp.MacroList;
import pmp.Macroprocessor;
import pmp.io.FileIO;
import pmp.xml.XMLConfig;

/**
 * Task for Ant.
 * <p>
 * Title: PMP: Macroprocessor
 * </p>
 * <p>
 * Description: Java macroprocessor
 * </p>
 * <p>
 * Copyright: Copyright (c) 2005
 * </p>
 * 
 * @author Lud\u011Bk Hlav\u00E1\u010Dek
 * @version 1.0
 */

public class PMPTask extends MatchingTask {

    /**
     * encoding to convert from
     */
    private String srcEncoding = null;

    /**
     * encoding to convert to
     */
    private String destEncoding = null;

    /**
     * debug level for macroprocessor
     */
    private String debugLevel = null;

    /**
     * Where to find input files
     */
    private File srcDir = null;

    /**
     * Where to put output files
     */
    private File destDir = null;

    /**
     * Prefix for imported properties
     */
    private String importPrefix = "ant:";

    @SuppressWarnings("unused")
    private FileUtils fileUtils = FileUtils.newFileUtils();

    private boolean defaultConfig = true;

    private boolean failonerror = true;

    private Mapper mapper = null;

    private ConfigElement config = null;

    private ExportElement export = null;

    private XMLConfig configXml = null;

    private Macroprocessor mp = null;

    private XMLConfig defXML = null;

    public PMPTask()
    {
        // no initialization required
    }

    /**
     * Set debug level.
     * 
     * @see Macroprocessor#setDebugLevel(String)
     * 
     * @param level
     *            desired debug level
     */
    public void setDebug(String debug)
    {
        this.debugLevel = debug;
    }

    /**
     * The value of the debug attribute.
     * 
     */
    public String getDebug()
    {
        return this.debugLevel;
    }

    /**
     * Set the prefix for importing Ant properties.
     * 
     * @param prefix
     *            Prefix to be prebended befor property name befor importing to
     *            macroprocessor.
     * 
     */
    public void setImportPrefix(String prefix)
    {
        this.importPrefix = prefix;
    }

    /**
     * The value of the importPrefix attribute.
     * 
     */
    public String getImportPrefix()
    {
        return this.importPrefix;
    }

    /**
     * Set the encoding to translate from. If unset, the default encoding for
     * the JVM is used.
     * 
     * @param encoding
     *            String containing the name of the Native encoding to convert
     *            from.
     */
    public void setSrcEncoding(String encoding)
    {
        this.srcEncoding = encoding;
    }

    /**
     * The value of the srcEncoding attribute.
     * 
     */
    public String getSrcEncoding()
    {
        return srcEncoding;
    }

    /**
     * Set the encoding to translate to. If unset, the default encoding for the
     * JVM is used.
     * 
     * @param encoding
     *            String containing the name of the Native encoding to convert
     *            to.
     */
    public void setDestEncoding(String encoding)
    {
        this.destEncoding = encoding;
    }

    /**
     * The value of the destEncoding attribute.
     * 
     */
    public String getDestEncoding()
    {
        return destEncoding;
    }

    /**
     * Set value of failonerror attribute. If <code>true</code>,
     * macroprocessor wil terminate on first error. Otherwise remaining files
     * will be processed. Default value is <code>true</code>.
     * 
     * @param config .
     */
    public void setFailOnError(boolean fail)
    {
        this.failonerror = fail;
    }

    /**
     * The value of the failOnError attribute.
     * 
     */
    public boolean getFailOnError()
    {
        return this.failonerror;
    }

    /**
     * Set value of defaultConfig property. If <code>true</code>,
     * macroprocessor wil load default configuration. If unset, the value
     * depends on presence of <code>config</code> element.
     * 
     * @param config .
     */
    public void setDefaultConfig(boolean config)
    {
        this.defaultConfig = config;
    }

    /**
     * The value of the defaultConfig attribute.
     * 
     */
    public boolean getDefaultConfig()
    {
        return defaultConfig;
    }

    /**
     * Set the source directory in which to find files to convert.
     * 
     * @param srcDir
     *            directory to find input file in.
     */
    public void setSrc(File srcDir)
    {
        this.srcDir = srcDir;
    }

    /**
     * Set the destination directory to place converted files into.
     * 
     * @param destDir
     *            directory to place output file into.
     */
    public void setDest(File destDir)
    {
        this.destDir = destDir;
    }

    /**
     * Defines the FileNameMapper to use (nested mapper element).
     * 
     * @return the mapper to use for file name translations.
     * 
     * @throws BuildException
     *             if more than one mapper is defined.
     */
    public Mapper createMapper() throws BuildException
    {
        if (mapper != null) { throw new BuildException(
                "Cannot define more than one mapper", getLocation()); }
        mapper = new Mapper(getProject());
        return mapper;
    }

    public ConfigElement createConfig()
    {
        if (config != null) { throw new BuildException(
                "Only one \"config\" element is allowed"); }
        config = new ConfigElement();
        return config;
    }

    public ExportElement createExport()
    {
        if (export != null) { throw new BuildException(
                "Only one \"export\" element is allowed"); }
        export = new ExportElement();
        return export;
    }

    /**
     * Imports properties before parsing file.
     */
    private void importProperties()
    {
        Hashtable props = getProject().getProperties();
        Enumeration keys = props.keys();
        MacroList ml = mp.getMPMacros();

        while (keys.hasMoreElements()) {
            Object key = keys.nextElement();
            String keyname = importPrefix + key.toString();
            ml.setMacro(keyname, props.get(key).toString());
        }// while
    }

    /**
     * Update export data. Performed after each processed file.
     */
    private void updateExports()
    {
        if (export == null) return;
        MacroList ml = mp.getMPMacros();
        for (ExportElement.ExportItem e : export.exports) {

            // ALL is processed in doExport()
            if (e.type == ExportElement.TYPE_ALL) continue;

            AbstractMacro m = ml.getMacro(e.macro);
            String mval = null;
            if (m != null) {
                switch (e.type) {
                case ExportElement.TYPE_MACRO:
                    mval = m.toString();
                    break;
                case ExportElement.TYPE_CLASSNAME:
                    mval = m.getClass().getName();
                    break;
                case ExportElement.TYPE_ALL:
                default:
                    // should never happen
                    assert false;
                }// switch
            }// if

            switch (e.mode) {
            case FIRST:
                if ((e.value == null) && (m != null)) e.value = mval;
                break;
            case LAST:
                if (m != null) e.value = mval;
                break;
            case CONCAT:
                if (m != null) {
                    if (e.value == null)
                        e.value = mval;
                    else
                        e.value = e.value + mval;
                }// if
                break;
            case SUM:
                if (m != null) {
                    int val = 0;
                    try {
                        val = Integer.parseInt(mval);
                    } catch (NumberFormatException ex) {
                        /* ignore invalid values */
                    }

                    if (e.value == null)
                        e.value = String.valueOf(val);
                    else
                        e.value = String.valueOf(Integer.parseInt(e.value)
                                + val);
                }// if
                break;
            }// switch
        }// for
    }

    /**
     * Exports macros according to given rules.
     */
    private void doExport()
    {
        for (ExportElement.ExportItem e : export.exports) {
            if (e.type == ExportElement.TYPE_ALL) {
                MacroList ml = mp.getMPMacros();
                String[] macros = ml.getMacroNames();
                for (String macro : macros) {
                    // TODO: pattern matching
                    String val = ml.getMacro(macro).toString();
                    getProject().setNewProperty(macro, val);
                }// for
            }
            else if (e.value != null) {
                getProject().setNewProperty(e.name, e.value);
            }// if
        }// for
    }

    /**
     * Execute the task
     * 
     * @throws BuildException
     *             is there is a problem in the task execution.
     */
    @Override
    public void execute() throws BuildException
    {
        try {

            DirectoryScanner scanner = null; // Scanner to find our inputs
            String[] files; // list of files to process

            // default srcDir to basedir
            if (srcDir == null) {
                srcDir = getProject().resolveFile(".");
            }

            // Require destDir
            if (destDir == null) { throw new BuildException(
                    "The dest attribute must be set."); }

            // if src and dest dirs are the same, require the mapper
            // to be set, so we don't stomp every file. One could still
            // include a file with the same extension, but ....
            if (srcDir.equals(destDir) && mapper == null) { throw new BuildException(
                    "A mapper must be set if src and dest dirs are the same."); }

            FileNameMapper m = null;
            if (mapper == null) {
                m = new IdentityMapper();
            }
            else {
                m = mapper.getImplementation();
            }

            scanner = getDirectoryScanner(srcDir);
            files = scanner.getIncludedFiles();
            SourceFileScanner sfs = new SourceFileScanner(this);
            files = sfs.restrict(files, srcDir, destDir, m);
            int count = files.length;
            if (count == 0) { return; }
            String message = "Converting " + count + " file"
                    + (count != 1 ? "s" : "") + " from ";
            log(message + srcDir + " to " + destDir);

            mp = new Macroprocessor();
            FileIO io = new FileIO(mp, FileIO.STD_OUTPUT, true);
            mp.setDebugLevel(debugLevel);

            if ((!defaultConfig) && (config == null)) { throw new BuildException(
                    "Either defaultConfig must be true or config element must be present."); }

            if (defaultConfig) defXML = new XMLConfig(mp, true); // load default
            // config

            if (config != null) {
                configXml = new XMLConfig(mp, false);
                //configXml.reset();
                try {
                    config.build();
                } catch (ParserConfigurationException e) {
                    throw new BuildException("Cannot create XML document");
                }
                configXml.parseXMLDocument(config.xmldoc);
            }

            for (int i = 0; i < files.length; i++) {

                System.out.println(files[i] + " -> "
                        + m.mapFileName(files[i])[0]);

                mp.reset();
                try {
                    parse(files[i], m.mapFileName(files[i])[0]);

                } catch (Exception ex) {
                    String msg = ex.toString();
                    if (failonerror)
                        throw new BuildException(msg, getLocation());
                    else
                        System.err.println(msg);
                }

                io.closeFiles(false);
                updateExports();

            }// for

            if (export != null) {
                doExport();
            }
        } catch (NullPointerException ex) {
            ex.printStackTrace();
        }
    }

    /**
     * Convert a single file.
     * 
     * @param srcName
     *            name of the input file.
     * @param destName
     *            name of the input file.
     * @throws IOException
     */
    private void parse(String srcName, String destName) throws BuildException,
            IOException, FatalParserException
    {
        File srcFile; // File to convert
        File destFile; // where to put the results

        // Build the full file names
        srcFile = new File(srcDir, srcName);
        destFile = new File(destDir, destName);

        // Make sure we're not about to clobber something
        if (srcFile.equals(destFile)) { throw new BuildException("file "
                + srcFile + " would overwrite itself"); }

        File parentFile = destFile.getParentFile();
        if ((parentFile != null) && (!parentFile.exists())
                && (!parentFile.mkdirs()))
            throw new BuildException("cannot create parent directory for "
                    + destName);

        log("executing " + srcName, Project.MSG_VERBOSE);

        File temp = destFile;
        // File temp = fileUtils.createTempFile("temp", ".pmp", null);
        // temp.deleteOnExit();
        mp.setDebugLevel(debugLevel);

        MacroList ml = mp.getMPMacros();

        if (defaultConfig) {
            defXML.writeToMacroList(ml);
        }

        if (configXml != null) {
            configXml.writeToMacroList(ml);
        }

        importProperties();

        if (srcEncoding != null)
            ml.setMacro(ml.forName(ConfigDirectives.INPUT_ENCODING),
                    srcEncoding);
        if (destEncoding != null)
            ml.setMacro(ml.forName(ConfigDirectives.OUTPUT_ENCODING),
                    destEncoding);

        if (debugLevel != null) mp.setDebugLevel(debugLevel);

        mp.getMPIO().setOutput(temp.getAbsolutePath());
        mp.getMPIQ().pushBack(mp.getMPIO().getInput(srcFile.getAbsolutePath()));

        if (mp.isReady()) {
            mp.parse();
        }
        else {
            throw new BuildException("Couldn't initialize macroprocessor for "
                    + srcName, getLocation());
        }

        // io.closeFiles(false);
        // fileUtils.rename(temp, destFile);

    }

}
