package pmp.cli;

import java.util.*;

import pmp.*;
import pmp.io.FileIO;
import pmp.xml.XMLConfig;

import static pmp.cli.Launcher.*;

/**
 * Main class for CLI interface.
 * <p>
 * Title: PMP: Macroprocessor
 * </p>
 * <p>
 * Description: Java macroprocessor
 * </p>
 * <p>
 * Copyright: Copyright (c) 2005
 * </p>
 * 
 * @author Lud\u011Bk Hlav\u00E1\u010Dek
 * @version 1.0
 */

public class Main {

    /**
     * Main method. Invokes <code>run</code> on created instance of ths class.
     * 
     * @param args
     * @throws Exception
     */
    public static void main(String[] args) throws Exception
    {
        (new Main()).run(args);
    }

    private List<String> sources = new ArrayList<String>(16);

    /** List of macro definitions passed on command line. */
    private Map<String, AbstractMacro> macros = new Hashtable<String, AbstractMacro>(
            32);

    private void addMacroDefFromCLI(String def)
    {
        int d = def.indexOf('=');
        String d1 = def.substring(0, d);
        String d2 = def.substring(d + 1);
        macros.put(d1, new DefinedMacro(d2));
    }

    private void addMacroClassFromCLI(String def)
    {
        int d = def.indexOf('=');
        try {
            Class macroClass = Class.forName(def.substring(d + 1));
            macros.put(def.substring(0, d), (AbstractMacro) macroClass
                    .newInstance());
        } catch (Throwable t) {
            System.err.println("Could not load class " + d + ".");
            System.err.println("Reason: " + t.toString());
        }
    }

    /**
     * Macro definition.
     */
    public static final String OPTION_DEFINE = "-D";

    /**
     * Load macro from class.
     */
    public static final String OPTION_BUILTIN = "-m";

    /**
     * Debug level.
     */
    public static final String OPTION_DEBUG = "-d";

    /**
     * Configuration file.
     */
    public static final String OPTION_CONFIGFILE = "-c";

    /**
     * Output file.
     */
    public static final String OPTION_OUTPUTFILE = "-o";

    /**
     * Input encoding.
     */
    public static final String OPTION_INPUT_ENCODING = "-ie";

    /**
     * Output encoding.
     */
    public static final String OPTION_OUTPUT_ENCODING = "-oe";

    /**
     * End of options is signalized by <code>"--"</code>
     */
    public static final String END_OF_OPTIONS = "--";

    /**
     * Command line version of PMP. Parses arguments, creates instance of
     * macroprocessor and processes input files.
     * 
     * @param args
     *            command line arguments
     * @throws Exception
     */
    public void run(String[] args) throws Exception
    {
        // parse args
        int i = 0;
        boolean iargs = true;
        boolean customConfig = false;
        String outputFile = null;

        Macroprocessor mp = new Macroprocessor(); // create MP

        FileIO io = new FileIO(mp, InputOutput.STD_OUTPUT, true);

        MacroList ml = mp.getMPMacros();

        while (i < args.length) { // process parameters
            if (!iargs) {
                sources.add(args[i]);
            }
            else {
                try {
                    if (OPTION_OUTPUTFILE.equals(args[i])) {
                        outputFile = args[++i];

                    }
                    else if (OPTION_DEFINE.equals(args[i])) {
                        addMacroDefFromCLI(args[++i]);

                    }
                    else if (OPTION_BUILTIN.equals(args[i])) { // load class
                        addMacroClassFromCLI(args[++i]);

                    }
                    else if (OPTION_INPUT_ENCODING.equals(args[i])) {
                        ml.setMacro(ml.forName(
                                ConfigDirectives.INPUT_ENCODING), args[++i]);

                    }
                    else if (OPTION_OUTPUT_ENCODING.equals(args[i])) {
                        ml.setMacro(ml.forName(
                                ConfigDirectives.OUTPUT_ENCODING), args[++i]);
                        // force reopening of output
                        outputFile = InputOutput.STD_OUTPUT;

                    }
                    else if (OPTION_DEBUG.equals(args[i])) {
                        mp.setDebugLevel(args[++i]);

                    }
                    else if (OPTION_CONFIGFILE.equals(args[i])) {
                        String configFile = args[++i];
                        XMLConfig xml = new XMLConfig(mp, configFile);
                        xml.writeToMacroList(ml);
                        customConfig = true;

                    }
                    else if (OPTION_HELP.equals(args[i])) {
                        printHelp();
                        return;

                    }
                    else if (OPTION_VERSION.equals(args[i])) {
                        printHelp();
                        return;

                    }
                    else if (OPTION_GUI.equals(args[i])) { // end of options
                        System.err.println("Option \"" + OPTION_GUI
                                + "\" must be given as first parameter.");
                        return;

                    }
                    else if (END_OF_OPTIONS.equals(args[i])) { // end of
                                                                // options
                        iargs = false;

                    }
                    else if ((args[i].length() > 0)
                            && (args[i].charAt(0) == '-')) {
                        System.err.println("Unknown option: " + args[i]);
                        return;

                    }
                    else {
                        iargs = false;
                        i--;
                    }
                } catch (ArrayIndexOutOfBoundsException ex) {
                    System.err.println("Missing argument.");
                    System.err
                            .println("Try \"java -jar pmp.jar -help\" for syntax details.");
                    return;
                }
            } // if/else
            i++;
        } // while

        if (outputFile != null) {
            io.setOutput(outputFile);
        }
        if (!customConfig) {
            XMLConfig defXML = new XMLConfig(mp, true); // load default config
            defXML.writeToMacroList(ml);
        }

        ml.addAll(macros); // add user definitions

        InputQueue iq = mp.getMPIQ();

        if (sources.size() > 0) {
            for (int index = 0; index < sources.size(); index++) {
                iq.pushBack(io.getInput(sources.get(index)));
            } // for
        }// if
        else { // if no source files, use console input
            iq.pushBack(io.getInput(InputOutput.STD_INPUT));
        }

        if (mp.isReady()) {
            try {
                mp.parse();
            } catch (FatalParserException ex) {
                System.err.println("Parser terminated with error: "
                        + ex.getMessage());
            }
            io.closeFiles(true);
        }
        else {
            System.err.println("Couldn't initialize macroprocessor");
        }
    }

}
