package pmp.gui;

import java.awt.*;
import java.awt.event.*;
import java.util.Arrays;

import javax.swing.*;
import javax.swing.border.BevelBorder;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import pmp.*;
import pmp.io.StringIO;
import pmp.xml.XMLConfig;
import static pmp.gui.Const.*;
import static pmp.gui.GUITools.*;

/**
 * Main panel shared by GUI and Applet.
 * <p>
 * Title: PMP: Macroprocessor
 * </p>
 * <p>
 * Description: Java macroprocessor
 * </p>
 * <p>
 * Copyright: Copyright (c) 2005
 * </p>
 * 
 * @author Lud\u011Bk Hlav\u00E1\u010Dek
 * @version 1.0
 */

public class MainPanel extends JPanel implements ActionListener,
        MacroListModifyListener, ListSelectionListener {

    protected Macroprocessor mp;

    protected XMLConfig xml;

    private Container mainWindow;

    private ActionListener parentListener;

    private JToolBar toolBar;

    private JTabbedPane jtab;

    private JMenuItem menuReloadMacros;

    private JMenuItem menuShowErrorDialog;

    private JTextPane sourcePanel;

    private JEditorPane outputPanel;

    private JList macroList;

    private JLabel macroNameLabel;

    private JTextPane macroEdit;

    private JButton macroChangeButton;

    private DefaultListModel mpOutput;

    private int errors;

    private int warnings;

    private int notices;

    public MainPanel(Container parentContainer, ActionListener parentListener)
    {
        this.mp = new Macroprocessor();
        this.mp.assignIO(new StringIO(mp, ""));
        this.mainWindow = parentContainer;
        this.parentListener = parentListener;

        xml = new XMLConfig(mp, true); // load default config
        xml.writeToMacroList(mp.getMPMacros());

        createComponents();
        updateList();
    }

    /**
     * Loads XML config file.
     */
    private void loadXML()
    {
        // Create a file chooser
        final JFileChooser fc = new JFileChooser();
        // fc.setFileFilter(GUITools.getXmlFilter());
        int returnVal = fc.showOpenDialog(this);
        if (returnVal == JFileChooser.APPROVE_OPTION) {
            String configFile = fc.getSelectedFile().getAbsolutePath();
            xml = new XMLConfig(mp, configFile);
            mp.getMPMacros().clear(true);
            xml.writeToMacroList(mp.getMPMacros());
            updateList();
            JOptionPane.showMessageDialog(this, configFile + " loaded",
                    "Configuration file", JOptionPane.INFORMATION_MESSAGE,
                    GUITools.getIcon("loadxml"));
        }// if
    }

    /**
     * Creates toolbar button.
     * 
     * @param title
     *            button title
     * @param icon
     *            button icon (without path and extension)
     * @return JButton object
     */
    private JButton createToolButton(String title, String icon)
    {
        JButton button = new JButton(title);
        button.addActionListener(this);
        if (title != null) {
            button.setIcon(GUITools.getIcon(icon));
        }
        return button;
    }

    private JToolBar createToolBar()
    {
        toolBar = new JToolBar("Tool bar", JToolBar.HORIZONTAL);
        JButton btnRun = createToolButton(MENU_RUN, null);
        btnRun.setMnemonic(KeyEvent.VK_R);
        toolBar.add(btnRun);
        toolBar.setFloatable(false);
        toolBar.setPreferredSize(new Dimension(450, 40));
        return toolBar;
    }

    /**
     * Creates and shows main panel.
     */
    private void createComponents()
    {
        jtab = new JTabbedPane(JTabbedPane.TOP);

        JSplitPane mainSplitPane;
        // PROGRAM
        mainSplitPane = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
        sourcePanel = new JTextPane();
        sourcePanel.setPreferredSize(new Dimension(300, 100));
        sourcePanel.addKeyListener(new KeyAdapter() {

            @Override
            public void keyReleased(KeyEvent e)
            {
                if (!e.isControlDown()) return;
                if (e.getKeyChar() == 'V') {
                    sourcePanel.paste();
                }
                else if (e.getKeyChar() == 'C') {
                    sourcePanel.copy();
                }
                else if (e.getKeyChar() == 'X') {
                    sourcePanel.cut();
                }
            }
        });
        outputPanel = new JTextPane();
        outputPanel.setPreferredSize(new Dimension(300, 100));

        mainSplitPane.setTopComponent(new JScrollPane(sourcePanel));
        mainSplitPane.setBottomComponent(new JScrollPane(outputPanel));
        jtab.addTab(TAB_PROGRAM, mainSplitPane);

        // MACROS
        mainSplitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
        macroEdit = new JTextPane();
        macroEdit.setBorder(BorderFactory.createLineBorder(SystemColor.control,
                5));
        macroEdit.setPreferredSize(new Dimension(300, 300));

        macroList = new JList();
        macroList.setBorder(BorderFactory
                .createBevelBorder(BevelBorder.LOWERED));
        macroList.addListSelectionListener(this);

        JScrollPane listScroll = new JScrollPane(macroList);
        listScroll
                .setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED);
        listScroll
                .setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        listScroll.setPreferredSize(new Dimension(100, 400));

        JPanel editorPanel = new JPanel(new BorderLayout(4, 4));
        JPanel editorBar = new JPanel(new FlowLayout(FlowLayout.LEFT));
        editorBar.add(macroChangeButton = new JButton(BUTTON_CHANGE));
        macroChangeButton.setEnabled(false);
        macroChangeButton.addActionListener(this);
        editorBar.add(macroNameLabel = new JLabel(""));
        editorPanel.add(editorBar, BorderLayout.NORTH);
        editorPanel.add(macroEdit, BorderLayout.CENTER);

        mainSplitPane.setTopComponent(listScroll);
        mainSplitPane.setBottomComponent(editorPanel);

        jtab.addTab(TAB_MACROS, mainSplitPane);

        // OUTPUT
        mpOutput = new DefaultListModel();
        JList outList = new JList();
        outList.setModel(mpOutput);
        JScrollPane outputScroll = new JScrollPane(outList);
        listScroll
                .setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED);
        listScroll
                .setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        listScroll.setPreferredSize(new Dimension(400, 400));
        jtab.addTab(TAB_OUTPUT, outputScroll);

        // MAIN PANEL
        this.setLayout(new BorderLayout());
        this.add(createToolBar(), BorderLayout.NORTH);
        this.add(jtab, BorderLayout.CENTER);
    }

    /**
     * Creates menu item.
     * 
     * @param itemType
     *            type of menu item
     * @param title
     *            menu item title
     * @param icon
     *            menu item icon (without path and extension)
     * @param acc
     *            accelerator (KeyStroke)
     * @return JMenuItem object
     * @see pmp.gui.MainPanel.MenuType
     */
    protected JMenuItem createMenuItem(MenuType itemType, String title,
            String icon, KeyStroke acc)
    {
        JMenuItem item;
        switch (itemType) {
        case ITEM:
            item = new JMenuItem(title);
            break;
        case CHECKBOX:
            item = new JCheckBoxMenuItem(title);
            break;
        case RADIO:
            item = new JRadioButtonMenuItem(title);
            break;
        default:
            throw new InternalMPError(0x820186); 
        }
        item.setIcon(getIcon(icon));
        if (acc != null) {
            item.setAccelerator(acc);
        }
        item.addActionListener(this);
        return item;
    }

    enum MenuType {
        ITEM, CHECKBOX, RADIO
    }

    protected JMenuItem createMenuItem(String title, String icon, KeyStroke acc)
    {
        return createMenuItem(MenuType.ITEM, title, icon, acc);
    }

    /**
     * Creates main menu bar for main frame.
     * 
     * @return created menubar
     */
    public JMenuBar createMenu()
    {
        JMenuBar menubar = new JMenuBar();
        JMenu menu;

        menu = new JMenu(MENUBAR_PROGRAM); // File
        menu.setMnemonic(KeyEvent.VK_P);
        menu.add(createMenuItem(MENU_RUN, ICON_RUN, KeyStroke.getKeyStroke(
                KeyEvent.VK_R, InputEvent.ALT_MASK)));
        menu.add(new JSeparator());
        menu.add(createMenuItem(MENU_QUIT, ICON_EXIT, KeyStroke.getKeyStroke(
                KeyEvent.VK_X, InputEvent.ALT_MASK)));
        menubar.add(menu);

        menu = new JMenu(MENUBAR_MACROS); // Macros
        menu.setMnemonic(KeyEvent.VK_M);

        menu.add(createMenuItem(MENU_LOAD_DEF, ICON_LOADDEF, KeyStroke
                .getKeyStroke(KeyEvent.VK_D, InputEvent.ALT_MASK)));

        menu.add(createMenuItem(MENU_LOAD_XML, ICON_LOADXML, KeyStroke
                .getKeyStroke(KeyEvent.VK_L, InputEvent.ALT_MASK)));

        menu.add(new JSeparator());
        // autoload xml config
        menuReloadMacros = createMenuItem(MenuType.CHECKBOX,
                MENU_RELOAD_MACROS, ICON_EMPTY, KeyStroke.getKeyStroke(
                        KeyEvent.VK_N, InputEvent.ALT_MASK));
        menuReloadMacros.setSelected(true);
        menu.add(menuReloadMacros);

        menubar.add(menu);

        menu = new JMenu(MENUBAR_OPTIONS); // Options
        menu.setMnemonic(KeyEvent.VK_O);
        menuShowErrorDialog = createMenuItem(MenuType.CHECKBOX, MENU_SHOW_ERROR_DIALOG,
                ICON_RUN, KeyStroke.getKeyStroke(KeyEvent.VK_D,
                        InputEvent.ALT_MASK));
        menuShowErrorDialog.setSelected(true);
        menu.add(menuShowErrorDialog);
        menubar.add(menu);

        menu = new JMenu(MENUBAR_HELP); // Help
        menu.setMnemonic(KeyEvent.VK_H);
        menu.add(createMenuItem(MENU_ABOUT, ICON_ABOUT, KeyStroke.getKeyStroke(
                KeyEvent.VK_F1, KeyEvent.CTRL_MASK)));
        menu.add(createMenuItem(MENU_HELP, ICON_HELP, KeyStroke.getKeyStroke(
                KeyEvent.VK_F1, 0)));
        menubar.add(menu);

        return menubar;
    }

    /**
     * ActionListener for macroprocessor messages.
     */
    private ActionListener mpOutputListener = new ActionListener() {

        public void actionPerformed(ActionEvent e)
        {
            String text = e.getActionCommand();
            switch (e.getID()) {
            case StringIO.ERROR_EVENT:
                text = "Error: " + text;
                errors++;
                break;
            case StringIO.WARNING_EVENT:
                text = "Warning: " + text;
                warnings++;
                break;
            case StringIO.NOTICE_EVENT:
                text = "Notice: " + text;
                notices++;
                break;
            case StringIO.DEBUG_EVENT:
                text = "DEBUG: " + text;
                break;
            default:
                throw new InternalMPError(0x820243);
            }
            mpOutput.addElement(text);
        }
    };

    private void runMP()
    {
        errors = warnings = notices = 0;
        StringIO io = new StringIO(mp, sourcePanel.getText());
        io.addActionListener(mpOutputListener);
        mp.assignIO(io);


        if (menuReloadMacros.isSelected()) {
            mp.getMPMacros().clear(true);
            xml.writeToMacroList(mp.getMPMacros());
            // reload macros
        }
        if (mp.isReady()) {
            try {
            mp.parse();
            } catch (FatalParserException ex) {
                // parser failed
            }
            outputPanel.setText(io.getOutputBuilder().toString());
            if ((errors > 0) && menuShowErrorDialog.isSelected()) {
                JOptionPane.showMessageDialog(mainWindow, "Number of errors: " + errors + "\nSee \"Errors\" tab for details.");
            }
        }
        else {
            outputPanel.setText("Couldn't run Macroprocessor.");
        }
    }

    private void updateList()
    {
        updateScheduled = false;
        MacroList macros = mp.getMPMacros();
        String[] list = macros.getMacroNames();
        Arrays.sort(list);
        macroList.setListData(list);
    }

    private boolean updateScheduled = false;

    /**
     * This method is invoked when change in the macro list occurs.
     * 
     * @param event
     *            event describing modification
     */
    public void macroListModified(MacroListModifyEvent event)
    {
        if (!updateScheduled) {
            updateScheduled = true;
            SwingUtilities.invokeLater(macroListDispatcher);
        } // if
    }

    /**
     * ActionListener for macroprocessor messages.
     */
    private Runnable macroListDispatcher = new Runnable() {

        /**
         * This method is invoked by AWT thread after main thread receives
         * notification about macro list change.
         */
        public void run()
        {
            updateList();
        }
    };

    /**
     * Menu events dispatcher.
     * 
     * @param e
     *            menu/button event
     */
    public void actionPerformed(ActionEvent e)
    {
        String action = e.getActionCommand();

        if (Const.MENU_RUN.equals(action)) {
            runMP();
        }
        
        else if (MENU_RELOAD_MACROS.equals(action)) {
            // checked state is changed automaticaly
            // menuReloadMacros.setSelected(!menuReloadMacros.isSelected());
        }

        else if (MENU_LOAD_XML.equals(action)) {
            loadXML();
        }

        else if (MENU_LOAD_DEF.equals(action)) {
            //loadXML();
        }

        else if (BUTTON_CHANGE.equals(action)) {
            String macroName = macroNameLabel.getText();
            if (mp.getMPMacros().isReadOnly(macroName)) {
                JOptionPane.showMessageDialog(mainWindow, "Macro " + macroName
                        + " is read-only.");
            } // if
            else {
                mp.getMPMacros().setMacro(macroName, macroEdit.getText());
            }
        }

        else { // send to parent (QUIT, HELP, ABOUT, ...)
            parentListener.actionPerformed(e);
        }
    }

    /**
     * Listener for MacroList selection changes.
     * 
     * @param e
     *            event object
     */
    public void valueChanged(ListSelectionEvent e)
    {
        Object selected = macroList.getSelectedValue();
        if (selected != null) {
            String macroName = selected.toString();
            AbstractMacro macro = mp.getMPMacros().getMacro(macroName);
            if (macro instanceof DefinedMacro) {
                this.macroNameLabel.setText(macroName);
                this.macroEdit.setText(macro.toString());
                this.macroEdit.setEditable(true);
                this.macroChangeButton.setEnabled(true);
            }
            else {
                this.macroNameLabel.setText(macroName);
                this.macroEdit.setText(macro.getClass().getName());
                this.macroEdit.setEditable(false);
                this.macroChangeButton.setEnabled(false);
            }
        } // if
    }

    static final long serialVersionUID = -1L;

}
