package pmp.io;

import java.io.*;

import pmp.*;
import static pmp.ConfigDirectives.*;

/**
 * Implements filesystem InputOutput.
 * 
 * <p>
 * Title: PMP: Macroprocessor
 * </p>
 * <p>
 * Description: Java macroprocessor
 * </p>
 * <p>
 * Copyright: Copyright (c) 2005
 * </p>
 * 
 * @author Lud\u011Bk Hlav\u00E1\u010Dek
 * @version 1.0
 */

public class FileIO extends InputOutput {

   /**
    * InputSource class using file I/O.
    * <p>
    * Title: PMP: Macroprocessor
    * </p>
    * <p>
    * Description: Java macroprocessor
    * </p>
    * <p>
    * Copyright: Copyright (c) 2005
    * </p>
    * <p>
    * Company:
    * </p>
    * 
    * @author Luděk Hlaváček
    * @version 1.0
    */
   private class ReaderInput implements InputSource {

      private final Reader r;

      // private boolean feof = false;
      public ReaderInput(Reader r)
      {
         this.r = r;
      }

      public String read(int length)
      {
         char[] buffer = new char[length];
         try {
            int l = r.read(buffer);
            if (l == -1) {
               // feof = true;
               r.close();
               return "";
            }
            return new String(buffer, 0, l);

         } catch (IOException ex) {
            warning("Read error");
         }
         return "";
      }

      public void close()
      {
         try {
            r.close();
         } catch (IOException ex) {
            notice("Close failed");
         }
      }

   } // FileInput

   /**
    * InputSource class using standard I/O.
    * <p>
    * Title: PMP: Macroprocessor
    * </p>
    * <p>
    * Description: Java macroprocessor
    * </p>
    * <p>
    * Copyright: Copyright (c) 2005
    * </p>
    * <p>
    * Company:
    * </p>
    * 
    * @author Luděk Hlaváček
    * @version 1.0
    */
   private class ConsoleInput implements InputSource {

      private Reader r;

      private boolean feof = false;

      public ConsoleInput()
      {
         r = new InputStreamReader(System.in);
      }

      public String read(int length)
      {
         if (feof)
            return "";

         try {
            int l = r.read();
            if (l == -1) {
               feof = true;
               r.close();
               return "";
            }

            return String.valueOf((char) l);
         } catch (IOException ex) {
            warning("Read error");
         }
         return "";
      }

      public void close()
      {
         // do nothing.
      }

   }

   /**
    * Writer class simulating null device.
    * <p>
    * Title: PMP: Macroprocessor
    * </p>
    * <p>
    * Description: Java macroprocessor
    * </p>
    * <p>
    * Copyright: Copyright (c) 2005
    * </p>
    * <p>
    * Company:
    * </p>
    * 
    * @author Luděk Hlaváček
    * @version 1.0
    */
   private class NullOutputStream extends OutputStream {

      @Override
      public void write(int i)
      {
         // do nothnig
      }

   }

   /** Current output */
   protected Writer output;

  
   /**
    * Should be <code>output</code> closed when nott needed. Set to
    * <code>false</code> when using stdio.
    */
   protected boolean closeableOut;

   public FileIO(Macroprocessor mp, String outputFile, boolean assign)
   {
      super(mp);
      setOutput(outputFile);
      if (assign) mp.assignIO(this);
   }

   @Override
   public InputSource getInput(String inputFile)
   {
      if (STD_INPUT.equals(inputFile))
         return new ConsoleInput();

      if (NULL_INPUT.equals(inputFile))
         return new StringInputSource("");

      if (RANDOM_INPUT.equals(inputFile))
         return new RandomInputSource(null, -1);

      try {
         InputStream is = new FileInputStream(inputFile);
         if (getMPMacros().isDirectiveDefined(INPUT_ENCODING)) {
            String enc = getMPMacros().getString(INPUT_ENCODING);
            try {
               return new ReaderInput(new InputStreamReader(is, enc));
            } catch (UnsupportedEncodingException e) {
               error("Unsupported encoding: " + enc);
               return new ReaderInput(new InputStreamReader(is));
            }
         } else {
            return new ReaderInput(new InputStreamReader(is));
         }
      } catch (FileNotFoundException ex) {
         warning("File \"" + inputFile + "\" not found");
         return null;
      }
   }

   /**
    * This class doesn't provide default input source.
    * 
    * @return <code>null</code>
    */
   @Override
   public InputSource getDefaultInput()
   {
      return null;
   }

   @Override
   public void setOutput(String outputFile)
   {
      Writer prev = output;
      boolean prevclose = closeableOut;

      OutputStream os;

      if ("".equals(outputFile) || STD_OUTPUT.equals(outputFile)) {
         os = System.out;
         closeableOut = false;
      }

      else if (ERR_OUTPUT.equals(outputFile)) {
         os = System.err;
         closeableOut = false;
      }

      else if (NULL_OUTPUT.equals(outputFile)) {
         os = new NullOutputStream();
         closeableOut = true;
      }

      else {
         try {
            os = new FileOutputStream(outputFile);
            closeableOut = true;
         } catch (IOException ex) {
            System.err.println("Output redirection failed: " + ex.getMessage());
            os = System.out;
            closeableOut = false;
         }
      }

      if (getMPMacros().isDirectiveDefined(OUTPUT_ENCODING)) {
         String enc = getMPMacros().getString(OUTPUT_ENCODING);
         try {
            output = new OutputStreamWriter(os, enc);
         } catch (UnsupportedEncodingException e) {
            System.err.println("Unsupported encoding: " + enc);
            output = new OutputStreamWriter(os);
         }
      } else {
         output = new OutputStreamWriter(os);
      }

      try {
         if (prev != null)
             prev.flush();
          
         if ((prev != null) && prevclose) 
            prev.close();

      } catch (IOException ex) {
          // ignore exceptions on close
      }

      // write name of output into config directive "pmp:output"
      getMPMacros().setMacroForName(OUTPUT, outputFile);
   }

   /**
    * Writes given string into output writer.
    * 
    * @param str
    *           string to be written to output
    */
   @Override
   public void write(String str)
   {
      try {
         output.write(str);
      } catch (IOException ex) {
         error("IOException: " + ex.getMessage());
      }
   }

   /**
    * Close output file.
    * @param force close file even if it is marked as not closeable.
    */
   public void closeFiles(boolean force)
   {
      if (!force && !closeableOut)
         return;
      try {
         output.close();
      } catch (IOException ex) {
         notice("Couldn't close files.");
      }
   }

   @Override
   public void errorHandler(String error)
   {
      System.err.println(error);
   }

   @Override
   public void warningHandler(String warning)
   {
      System.err.println(warning);
   }

   @Override
   public void noticeHandler(String noticemessage)
   {
      System.err.println(noticemessage);
   }

   @Override
   public void debugHandler(String debugmessage)
   {
      System.err.println(debugmessage);

   }

}
