package pmp.io;

import java.util.*;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import pmp.*;

/**
 * Macroprocessor IO based on strings.
 * <p>
 * Title: PMP: Macroprocessor
 * </p>
 * <p>
 * Description: Java macroprocessor
 * </p>
 * <p>
 * Copyright: Copyright (c) 2005
 * </p>
 * 
 * @author Lud\u011Bk Hlav\u00E1\u010Dek
 * @version 1.0
 */

public class StringIO extends InputOutput {

   private final StringBuilder output;

   private InputSource in;

   private final String program;

   public StringIO(Macroprocessor mp, String program, StringBuilder output)
   {
      super(mp);
      this.output = output;
      this.program = program;
      in = new StringInputSource(program);
   }

   public StringIO(Macroprocessor mp, String program)
   {
      this(mp, program, new StringBuilder(program.length()));
   }

   public StringBuilder getOutputBuilder()
   {
      return this.output;
   }

   public String getProgram()
   {
      return this.program;
   }

   @Override
   public void write(String str)
   {
      output.append(str);
   }

   @Override
   public void setOutput(String name)
   {
      getMP().warning("Program tried to change output to: " + name);
   }

   @Override
   public InputSource getInput(String name)
   {
      getMP().warning("Program tried to read from: " + name);
      return new StringInputSource("");
   }

   @Override
   public InputSource getDefaultInput()
   {
      return in;
   }

   // ////////////////////////////////////////////////////////////
   // ////////////////////////////////////////////////////////////
   // //////////////////////////////////////////////////////////////

   public static final int ERROR_EVENT = ActionEvent.RESERVED_ID_MAX | 0x18F0FFFF;

   public static final int WARNING_EVENT = ActionEvent.RESERVED_ID_MAX | 0x18F1FFFF;

   public static final int NOTICE_EVENT = ActionEvent.RESERVED_ID_MAX | 0x18F2FFFF;

   public static final int DEBUG_EVENT = ActionEvent.RESERVED_ID_MAX | 0x18F3FFFF;

   private List<ActionListener> listeners = new ArrayList<ActionListener>(4);

   /**
    * Adds new listeners for message events from macroprocessor.
    * 
    * @param l
    *           ActionListener to be added
    * @return <code>true</code> if the object was added
    */
   public boolean addActionListener(ActionListener l)
   {
      if (!listeners.contains(l)) {
         listeners.add(l);
         return true;
      }
      return false;
   }

   /**
    * Removes given acion listener from list.
    * 
    * @param l
    *           ActionListener to be removed
    * @return <code>true</code> if the object was removed from list
    */
   public boolean removeActionListener(ActionListener l)
   {
      return listeners.remove(l);
   }

   private void fireActionEvent(int type, String msg)
   {
      ActionEvent e = new ActionEvent(this, type, msg);
      Iterator<ActionListener> it = listeners.iterator();
      while (it.hasNext()) {
         (it.next()).actionPerformed(e);
      } // while
   }

   @Override
   public void errorHandler(String error) throws FatalParserException
   {
      fireActionEvent(ERROR_EVENT, error);
   }

   @Override
   public void warningHandler(String warning)
   {
      fireActionEvent(WARNING_EVENT, warning);
   }

   @Override
   public void noticeHandler(String noticemessage)
   {
      fireActionEvent(NOTICE_EVENT, noticemessage);
   }

   @Override
   public void debugHandler(String debugmessage)
   {
      fireActionEvent(DEBUG_EVENT, debugmessage);
   }

}
