package pmp.macro;

import pmp.*;
import pmp.xml.*;

/**
 * Allows manipulation with list of macros. This macro modifies
 * {@link pmp.MacroList} component of {@link pmp.Macroprocessor}.
 * <p>
 * This macro exports or imports macro list according to specified action.
 * </p>
 * <p>
 * Title: PMP: Macroprocessor
 * </p>
 * <p>
 * Description: Java macroprocessor
 * </p>
 * <p>
 * Copyright: Copyright (c) 2005
 * </p>
 * 
 * @author Lud\u011Bk Hlav\u00E1\u010Dek
 * @version 1.0
 */
public class MacroTable extends AbstractMacro {

    public static final int PARAMS_REQUIRED = 2;

    /**
     * Exports current list of macros into file given in the second argument.
     */
    public static final String ACTION_EXPORT = "EXPORT";

    /**
     * Imports macros from given configuration file.
     * <p>
     * Previously undefined macros are added, existing macros are replaced and
     * others are kept unchanged.
     * </p>
     * <p>
     * Read-only macros are kept unchanged as well.
     * </p>
     */
    public static final String ACTION_IMPORT = "IMPORT";

    /**
     * If given in second parameter, this macro will load default configuration
     * instead of reading external file. Can be used with {@link #ACTION_IMPORT}
     * and {@link #ACTION_REPLACE}.
     */
    public static final String DEFAULT_CONFIG = "pmp://config.default";

    /**
     * Replaces macro list with given list.
     * <p>
     * All but read-only macros are deleted prior importing new definitions.
     * </p>
     */
    public static final String ACTION_REPLACE = "REPLACE";

    @Override
    public String run(Macroprocessor mp, String[] args)
    {
        MacroList ml = mp.getMPMacros();
        String arg1 = args[1].trim().toUpperCase();

        if (ACTION_EXPORT.equals(arg1)) {
            String result = XMLTools.exportConfig(ml.getAll(), args[2]);
            if (result != null)
                mp.warning("Macro MacroTable: export failed: " + result);
        }
        else if (ACTION_IMPORT.equals(arg1)) {
            XMLConfig xml;
            if (args[2].equalsIgnoreCase(DEFAULT_CONFIG))
                xml = new XMLConfig(mp, true);
            else
                xml = new XMLConfig(mp, args[2]);
            xml.writeToMacroList(ml);
        }
        else if (ACTION_REPLACE.equals(arg1)) {
            XMLConfig xml;
            if (args[2].equalsIgnoreCase(DEFAULT_CONFIG))
                xml = new XMLConfig(mp, true);
            else
                xml = new XMLConfig(mp, args[2]);
            ml.clear(false);
            xml.writeToMacroList(ml);
        }

        return "";
    }

}
