package pmp.macro.java;

import java.util.*;
import java.util.regex.Matcher;

import pmp.RegExps;

/**
 * Internal class for manipulation with objects.
 * <p>
 * This class keeps track of created objects. It is used by all macros from
 * package pmp.macro.java.
 * </p>
 * <p>
 * Objects are identified by string identifiers. See {@link #OBJECT_IDENTIFIER}.
 * </p>
 * <p>
 * Title: PMP: Macroprocessor
 * </p>
 * <p>
 * Description: Java macroprocessor
 * </p>
 * <p>
 * Copyright: Copyright (c) 2005
 * </p>
 * 
 * @author Luděk Hlaváček
 * @version 1.0
 */

final class ObjectManager {

    private List<Object> objects;

    private ObjectManager()
    {
        objects = new ArrayList<Object>(64);
        objects.add(null); // first item is null object
    }

    private static ObjectManager instance = null;

    /**
     * Returns instance of ObjectManager object
     * 
     * @return universal instance of ObjectManager
     */
    public static ObjectManager getInstance()
    {
        if (instance == null) {
            instance = new ObjectManager();
        } // if
        return instance;
    }

    static final char OBJ_ID_SEPARATOR = '@';

    static final String OBJ_ID_CLASSNAME = "([\\[a-zA-Z_][a-zA-Z0-9._$]*)";

    static final String OBJ_ID_INDEX = "([0-9]+)";

    /**
     * Pattern which matches object identifiers. Valid identifiers consist of
     * class name, character <code>"@"</code> and index in object list. Valid
     * identifier might look like <code>java.lang.Long&#64;15</code>.
     */
    static final String OBJECT_IDENTIFIER = OBJ_ID_CLASSNAME + OBJ_ID_SEPARATOR
            + OBJ_ID_INDEX;

    private final Matcher idMatcher = RegExps.compilePattern(OBJECT_IDENTIFIER)
            .matcher("");

    /**
     * Extracts class name from object identifier.
     * <p>
     * Note: Object identifier is not checked for validity.
     * </p>
     * 
     * @param object
     *            object identifier
     * @return class name
     */
    String extractClassName(String object)
    {
        idMatcher.reset(object);
        if (idMatcher.matches()) { return idMatcher.group(1); }
        throw new IllegalArgumentException("Invalid object identifier: "
                + object);
    }

    /**
     * Extracts object number from object identifier.
     * <p>
     * Note: Object identifier is not checked for validity.
     * </p>
     * 
     * @param object
     *            object identifier
     * @return class name extracted from object identifier
     */
    int extractObjectId(String object)
    {
        idMatcher.reset(object);
        if (idMatcher.matches()) { return Integer.parseInt(idMatcher.group(2)); }
        throw new IllegalArgumentException("Invalid object identifier: "
                + object);
    }

    /**
     * Converts array of object identifiers to array of appropriate Class
     * objects.
     * 
     * @param args
     *            list of object identifiers
     * @param start
     *            index of first identifier in <code>args</code>
     * @return array of classes
     * @throws ClassNotFoundException
     */
    Class[] getClasses(String[] args, int start) throws ClassNotFoundException
    {
        final Class[] result = new Class[args.length - start];
        for (int i = start; i < args.length; i++) {
            result[i - start] = Class.forName(extractClassName(args[i]));
        }// for
        return result;
    }

    /**
     * Map of primitives and associated wrapper classes.
     */
    static final Map<String, String> primitives = new HashMap<String, String>();

    static {
        primitives.put("char", "java.lang.Character");
        primitives.put("byte", "java.lang.Byte");
        primitives.put("short", "java.lang.Short");
        primitives.put("int", "java.lang.Integer");
        primitives.put("long", "java.lang.Long");
        primitives.put("float", "java.lang.Float");
        primitives.put("double", "java.lang.Double");
        primitives.put("boolean", "java.lang.Boolean");
    }

    /**
     * Returns <code>true</cod> if it is possible to call method
     * with formal parameter types given in <code>args</code>
     * and actual parameters given in <code>values</code>.
     * 
     * @param args  parameters accepted by the method
     * @param values values passed to the method
     * @return <code>true</code> if it's ok
     */
    static boolean testCompatibility(Class<?>[] args, Class<?>[] values)
    {
        if (args.length != values.length) return false;

        for (int i = 0; i < args.length; i++) {

            if (args[i].isPrimitive()) {
                String w = primitives.get(args[i].getName());
                assert (w != null);// void?
                if (!values[i].getName().equals(w)) return false;

            }
            else {
                if (!args[i].isAssignableFrom(values[i])) return false;
            }
        }// for

        return true;
    }

    /**
     * Fetches objects with given identifiers as Object array.
     * 
     * @param args
     *            array of object identifiers
     * @param start
     *            index of first object identifier
     * @return array of objects
     * @throws ClassNotFoundException
     */
    Object[] getObjects(String[] args, int start) throws ClassNotFoundException
    {
        final Object[] result = new Object[args.length - start];
        for (int i = start; i < args.length; i++) {
            result[i - start] = getObject(args[i]);
        }// for
        return result;
    }

    /**
     * Adds object to list.
     * 
     * @param object
     *            Object to be added. Must not be <code>null</code>.
     * @return object identifier
     * @see #OBJECT_IDENTIFIER
     */
    String addObject(Object object)
    {
        assert (object != null);

        objects.add(object);
        return object.getClass().getName() + "@"
                + Integer.toString(objects.size() - 1);
    }

    /**
     * Adds null object to list.
     * 
     * @param className
     *            Class of this null object.
     * @return object identifier
     * @see #OBJECT_IDENTIFIER
     */
    String addNullObject(String className) throws ClassNotFoundException
    {
        // test if the specified name is valid className
        Class.forName(className);
        objects.add(null);
        return className + "@" + Integer.toString(objects.size() - 1);
    }

    /**
     * Returns object with specified index.
     * 
     * @param objectId
     *            index of object to be returned
     * @return object retrieved from list or <code>null</code> when object not
     *         found
     */
    Object getObject(String objectId)
    {
        idMatcher.reset(objectId);
        if (idMatcher.matches()) {
            final int i = Integer.parseInt(idMatcher.group(2));
            final Object obj = objects.get(i);
            if ((obj != null)
                    && !idMatcher.group(1).equals(obj.getClass().getName()))
                throw new IllegalArgumentException("Invalid object class: "
                        + objectId);
            return obj;
        }
        else {
            throw new IllegalArgumentException("Invalid object identifier: "
                    + objectId);
        }
    }

}
