package pmp.macro.math;

import java.util.*;
import java.util.regex.Matcher;

import pmp.AbstractMacro;
import pmp.Macroprocessor;
import pmp.RegExps;

import static pmp.macro.math.Tools.*;

/**
 * Evaluates given expression.
 * <p>
 * Recognized elements are integers and real numbers, operators <code>+</code>,<code>-</code>,<code>*</code>,<code>/</code>,<code>%</code>
 * (modulo), <code>^</code> (power), <code>=</code> (compare for equality),
 * <code>&gt;</code> (compare - greater than), <code>&lt;</code> (compare -
 * less then) brackets <code>(</code>, <code>)</code> and functions.
 * </p>
 * <p>
 * Functions are identified by ther name (case-insensitively) and arguments
 * enclosed in brackets and separated by commas. See documentation of enum class
 * {@link pmp.macro.math.Eval.Function} for list of available functions.
 * </p>
 * <p>
 * Table of operator priorities: <table summary="Table of operator priorities">
 * <tr>
 * <td>^</td>
 * <td>highest priority - power</td>
 * </tr>
 * <tr>
 * <td><code>*</code>, <code>/</code>, <code>%</code></td>
 * <td>multiplicative operators</td>
 * </tr>
 * <tr>
 * <td><code>+</code>, <code>-</code></td>
 * <td>additive operators</td>
 * </tr>
 * <tr>
 * <td><code>=</code>, <code>&gt;</code>, <code>&lt;</code></td>
 * <td>comparative operators</td>
 * </tr>
 * </table>
 * </p>
 * <p>
 * Whitespace characters are ignored.
 * </p>
 * <p>
 * Title: PMP: Macroprocessor
 * </p>
 * <p>
 * Description: Java macroprocessor
 * </p>
 * <p>
 * Copyright: Copyright (c) 2005
 * </p>
 * 
 * @author Lud\u011Bk Hlav\u00E1\u010Dek
 * @version 1.0
 * @todo constants (pi, e, log e, ...) as zero param functions
 */

public class Eval extends AbstractMacro {

    public static final int PARAMS_REQUIRED = 1;

    @Override
    public String run(Macroprocessor mp, String[] args)
    {
        exp = Tools.removeSpaces(args[1]);
        if (exp.length() == 0) {
            mp.warning("Macro Eval: missing expression");
            return "";
        }// if

        try {
            Number res = parse(EndType.EXPR, Priority.PARENTHESIS);
            if (exp.length() == 0) return res.toString();
        } catch (IllegalArgumentException ex) {
            mp.warning("Macro Eval: " + ex.getMessage());
            return "";
        }

        mp.warning("Macro Eval: Invalid expression");
        return "";
    }

    public Eval()
    {
        hexMatcher = RegExps.compilePattern(NUMBER_FORMAT_HEX).matcher("");
        binMatcher = RegExps.compilePattern(NUMBER_FORMAT_BIN).matcher("");
        floatMatcher = RegExps.compilePattern(NUMBER_FORMAT_FLOAT).matcher("");
        strMatcher = RegExps.compilePattern(ID_FORMAT).matcher("");
    }

    protected static final String NUMBER_FORMAT_BIN = "0[bB]([01]+)";

    protected static final String NUMBER_FORMAT_HEX = "0[xX]([0-9A-Fa-f]+)";

    // this matches strings like "e1", but it doesn't matter
    protected static final String NUMBER_FORMAT_FLOAT = "([0-9]*((?:\\.[0-9]+)?(?:[Ee]-?[0-9]+)?))";

    protected static final String ID_FORMAT = "[_A-Za-z]+";

    private Matcher floatMatcher, binMatcher, hexMatcher, strMatcher;

    private String exp;

    private static class Token {

        enum Type {
            INT, FLOAT, OP, COMA, ID, EOS, INVALID
        }

        Type type;

        String id;

        Long intval;

        Double floatval;

        Token(Type t)
        {
            this.type = t;
        }

        String assign(Type t, Matcher m, String s)
        {
            int len;
            if (m.groupCount() == 0) {
                len = m.end();
                this.id = s.substring(0, len);
            }
            else {
                len = m.end(1);
                this.id = s.substring(m.start(1), len);
            }
            this.type = t;
            return s.substring(len);
        }

        boolean isOp(String op)
        {
            return ((this.type == Type.OP) && (this.id.equals(op)));
        }

        boolean isNumber()
        {
            return ((this.type == Type.FLOAT) || (this.type == Type.INT));
        }

        Number asNumber()
        {
            if (this.type == Type.FLOAT)
                return this.floatval;
            else if (this.type == Type.INT)
                return this.intval;
            else
                assert (false);
            return null;
        }

    }// Token

    private Token nextToken()
    {
        Token t = new Token(Token.Type.INVALID);
        if (exp.length() == 0) return new Token(Token.Type.EOS);
        char c0 = Character.toLowerCase(exp.charAt(0));

        if ((c0 >= '0') && (c0 <= '9')) { // number
            hexMatcher.reset(exp);
            if (hexMatcher.lookingAt()) {
                exp = t.assign(Token.Type.INT, hexMatcher, exp);
                t.intval = Long.valueOf(t.id, 16);
                return t;
            }// if
            binMatcher.reset(exp);
            if (binMatcher.lookingAt()) {
                exp = t.assign(Token.Type.INT, binMatcher, exp);
                t.intval = Long.valueOf(t.id, 2);
                return t;
            }// if
            floatMatcher.reset(exp);
            if ((floatMatcher.lookingAt())) {
                if (floatMatcher.group(2).length() > 0) {
                    exp = t.assign(Token.Type.FLOAT, floatMatcher, exp);
                    t.floatval = Double.valueOf(t.id);
                    return t;
                }
                else {
                    exp = t.assign(Token.Type.INT, floatMatcher, exp);
                    t.intval = Long.valueOf(t.id, 10);
                    return t;
                }// else
            }// if

        }
        else if (c0 == '.') { // float number
            floatMatcher.reset(exp);
            if (floatMatcher.lookingAt()) {
                exp = t.assign(Token.Type.FLOAT, floatMatcher, exp);
                t.floatval = Double.valueOf(t.id);
                return t;
            }// if

        }
        else if ((c0 >= 'a') && (c0 <= 'z')) {
            strMatcher.reset(exp);
            if (strMatcher.lookingAt()) {
                exp = t.assign(Token.Type.ID, strMatcher, exp);
                return t;
            }// if

        }
        else if (c0 == ',') {
            t.type = Token.Type.COMA;
            t.id = exp.substring(0, 1);
            exp = exp.substring(1);
            return t;

        }
        else if ("=><+-*/()%^".indexOf(c0) != -1) {
            t.type = Token.Type.OP;
            t.id = exp.substring(0, 1);
            exp = exp.substring(1);
            return t;
        }
        throw new IllegalArgumentException("Invalid token: " + exp);
    }

    /**
     * Priority classes.
     * Used by expression parser to maintain correct 
     * order of expression evaluation.
     */
    private enum Priority {
        POWER(30), MUL(20), ADD(10), CMP(5), PARENTHESIS(1);

        Priority(int p)
        {
            this.p = p;
        }

        int getPriority()
        {
            return p;
        }

        private int p;
    }// Priority

    /**
     * The way current parser should end.
     * 
     */
    private enum EndType {
        /** Parser should reach end of expression. */
        EXPR,
        /** Parser should encounter ")". */
        BRACKET,
        /** Parser should encounter "," or ")". */
        PARAM
    }

    private static final Short NO_MORE_PARAMS = new Short((short) 1);

    private Number evalFunction(Token fnc)
    {
        Token t = nextToken();
        if (!t.isOp("("))
            throw new IllegalArgumentException("Unexpected token: " + t.id);
        List<Number> args = new ArrayList<Number>(4);
        while (true) {
            Number n = parse(EndType.PARAM, Priority.PARENTHESIS);
            if (n.equals(NO_MORE_PARAMS)) break;
            args.add(n);
        }// while
        Function f = Function.valueOf(fnc.id.toUpperCase());
        return f.eval(args.toArray(new Number[0]));
    }

    /**
     * This enumerated type handles function calling. Most of functions are
     * wrappers for functions defined in {@link java.lang.Math} class.
     * 
     */
    protected enum Function {
        /**
         * Sums arbitrary number of numbers.
         */
        ADD(1, -1),
        /**
         * Subtracts two numbers.
         */
        SUB(2, 2),
        /**
         * Multiplies arbitrary number of numbers.
         */
        MUL(1, -1),
        /**
         * Performs integer division.
         */
        DIV(2, 2),
        /**
         * Returns the value of the first argument raised to the power of the
         * second argument.
         */
        POWER(2, 2),
        /**
         * Returns negation of given number.
         */
        NEG(1, 1),
        /**
         * Computes division modulo.
         */
        MOD(2, 2),
        /**
         * Returns the trigonometric sine of an angle.
         */
        SIN(1, 1),
        /**
         * Returns the trigonometric cosine of an angle.
         */
        COS(1, 1),
        /**
         * Returns the arc sine of an angle, in the range of -pi/2 through pi/2.
         */
        ASIN(1, 1),
        /**
         * Returns the arc cosine of an angle, in the range of 0.0 through pi.
         */
        ACOS(1, 1),
        /**
         * Returns the trigonometric tangent of an angle.
         */
        TAN(1, 1),
        /**
         * Returns the hyperbolic tangent of an angle.
         */
        TANH(1, 1),
        /**
         * Returns the arc tangent of an angle, in the range of -pi/2 through
         * pi/2.
         */
        ATAN(1, 1),
        /**
         * Returns Euler's number <i>e</i> raised to the power of a given
         * value.
         */
        EXP(1, 1),
        /**
         * Returns the natural logarithm (base <i>e</i>) of a given value.
         */
        LN(1, 1),
        /**
         * Returns the base 10 logarithm of a ven value.
         */
        LOG10(1, 1),
        /**
         * Returns sqrt(x<sup>2</sup>+y<sup>2</sup>).
         */
        HYPOT(2, 2),
        /**
         * Returns absolute value of given nuber.
         */
        ABS(1, 1),
        /**
         * Returns square root of a given value.
         */
        SQRT(1, 1),
        /**
         * Returns the cube root of a given value.
         */
        CBRT(1, 1),
        /**
         * Returns Euler's number <i>e</i> raised to the power of a given
         * value.
         */
        COSH(1, 1),
        /**
         * Returns maximum of the given values.
         */
        MAX(1, -1),
        /**
         * Returns minimum of the given values.
         */
        MIN(1, -1),
        /**
         * If value is given then integer greater than or equal to 0 and less
         * than given value is returned. Otherwise this function returns random
         * number greater than or equal to 0.0 and less than 1.0.
         * 
         */
        RANDOM(0, 1),
        /**
         * Returns the signum function of the argument; zero if the argument is
         * zero, 1.0 if the argument is greater than zero, -1.0 if the argument
         * is less than zero.
         */
        SIGNUM(1, 1),
        /**
         * Returns the closest integer to the argument.
         */
        ROUND(1, 1),
        /**
         * Returns the smallest (closest to negative infinity) integer that is
         * greater than or equal to the argument.
         */
        CEIL(1, 1),
        /**
         * Returns the largest (closest to positive infinity) integer that is
         * less than or equal to the argument.
         */
        FLOOR(1, 1),
        /**
         * Compares given numbers. Returns 1 if the arguments are equal,
         * otherwise returns zero.
         * 
         */
        CMP(2, 2),
        /**
         * Compares given numbers. Returns 1 if the first argument is greater
         * than the second otherwise returns zero.
         * 
         */
        GT(2, 2),
        /**
         * Compares given numbers. Returns 1 if the first argument is less than
         * the second otherwise returns zero.
         * 
         */
        LT(2, 2),
        /**
         * Compares given numbers. Returns 1 if the first argument is greater or
         * equal than the second otherwise returns zero.
         * 
         */
        GTE(2, 2),
        /**
         * Compares given numbers. Returns 1 if the first argument is less or
         * equal than the second otherwise returns zero.
         * 
         */
        LTE(2, 2);

        private final int min;

        private final int max;

        private Function(int min, int max)
        {
            this.min = min;
            this.max = max;
        }

        private Number eval(Number... args)
        {
            if ((this.min > args.length)
                    || ((this.max >= 0) && (this.max < args.length)))
                throw new IllegalArgumentException(
                        "Macro Eval: Invalid number of arguments for function "
                                + this.toString());
            switch (this) {
            case SIGNUM:
                return fncSignum(args[0]);
            case ROUND:
                return fncRound(args[0]);
            case CEIL:
                return fncCeil(args[0]);
            case FLOOR:
                return fncFloor(args[0]);
            case RANDOM:
                return fncRandom(args);
            case MAX:
                return fncMax(args);
            case MIN:
                return fncMin(args);
            case ADD:
                return fncSum(args);
            case SUB:
                return fncSub(args[0], args[1]);
            case MUL:
                return fncMul(args);
            case DIV:
                return fncDiv(args[0], args[1]);
            case POWER:
                return fncPower(args[0], args[1]);
            case NEG:
                return fncNeg(args[0]);
            case MOD:
                return fncMod(args[0], args[1]);
            case SIN:
                return fncSin(args[0]);
            case COS:
                return fncCos(args[0]);
            case COSH:
                return fncCosh(args[0]);
            case ASIN:
                return fncAsin(args[0]);
            case ACOS:
                return fncAcos(args[0]);
            case TAN:
                return fncTan(args[0]);
            case TANH:
                return fncTanh(args[0]);
            case ATAN:
                return fncAtan(args[0]);
            case EXP:
                return fncExp(args[0]);
            case SQRT:
                return fncSqrt(args[0]);
            case CBRT:
                return fncCbrt(args[0]);
            case HYPOT:
                return fncHypot(args[0], args[1]);
            case LN:
                return fncLn(args[0]);
            case LOG10:
                return fncLog10(args[0]);
            case ABS:
                return fncAbs(args[0]);
            case CMP:
                return fncCmp(args[0], args[1]);
            case GT:
                return fncGT(args[0], args[1]);
            case GTE:
                return fncGTE(args[0], args[1]);
            case LT:
                return fncGT(args[1], args[0]);
            case LTE:
                return fncGTE(args[1], args[0]);
            }// switch
            return null;
        }
    }

    private static Number fncSignum(Number arg)
    {
        return Long.valueOf(Math.round(Math.signum(arg.doubleValue())));
    }

    private static Number fncRound(Number arg)
    {
        return Long.valueOf(Math.round(arg.doubleValue()));
    }

    private static Number fncCeil(Number arg)
    {
        return Long.valueOf(Math.round(Math.ceil(arg.doubleValue())));
    }

    private static Number fncFloor(Number arg)
    {
        return Long.valueOf(Math.round(Math.floor(arg.doubleValue())));
    }

    private static Number fncRandom(Number... args)
    {
        if (args.length == 1) { return Long.valueOf(getRandom().nextLong()
                % args[1].longValue()); }
        return Double.valueOf(Math.random());
    }

    private static Number fncNeg(Number n)
    {
        if (n instanceof Long)
            return Long.valueOf(-n.longValue());
        else
            return Double.valueOf(-n.doubleValue());
    }

    private static Number fncSum(Number... args)
    {
        Number res = args[0];
        for (int i = 1; i < args.length; i++) {
            if ((res instanceof Long) && (args[i] instanceof Long))
                res = Long.valueOf(res.longValue() + args[i].longValue());
            else
                res = Double.valueOf(res.doubleValue() + args[i].doubleValue());
        }
        return res;
    }

    private static Number fncMax(Number... args)
    {
        Number res = args[0];
        for (int i = 1; i < args.length; i++) {
            if ((res instanceof Long) && (args[i] instanceof Long))
                res = Long.valueOf(Math.max(res.longValue(), args[i]
                        .longValue()));
            else
                res = Double.valueOf(Math.max(res.doubleValue(), args[i]
                        .doubleValue()));
        }
        return res;
    }

    private static Number fncMin(Number... args)
    {
        Number res = args[0];
        for (int i = 1; i < args.length; i++) {
            if ((res instanceof Long) && (args[i] instanceof Long))
                res = Long.valueOf(Math.min(res.longValue(), args[i]
                        .longValue()));
            else
                res = Double.valueOf(Math.min(res.doubleValue(), args[i]
                        .doubleValue()));
        }
        return res;
    }

    private static Number fncSub(Number arg0, Number arg1)
    {
        if ((arg0 instanceof Long) && (arg1 instanceof Long))
            return Long.valueOf(arg0.longValue() - arg1.longValue());
        else
            return Double.valueOf(arg0.doubleValue() - arg1.doubleValue());
    }

    private static Number fncMul(Number... args)
    {
        Number res = args[0];
        for (int i = 1; i < args.length; i++) {
            if ((res instanceof Long) && (args[i] instanceof Long))
                res = Long.valueOf(res.longValue() * args[i].longValue());
            else
                res = Double.valueOf(res.doubleValue() * args[i].doubleValue());
        }
        return res;
    }

    /** Integer division */
    private static Number fncDiv(Number arg0, Number arg1)
    {
        return Long.valueOf(arg0.longValue() / arg1.longValue());
    }

    /** TODO: assign function name ? */
    private static Number fncRealDiv(Number arg0, Number arg1)
    {
        if ((arg0 instanceof Long) && (arg1 instanceof Long))
            if ((arg0.longValue() % arg1.longValue()) == 0)
                return Long.valueOf(arg0.longValue() / arg1.longValue());

        return Double.valueOf(arg0.doubleValue() / arg1.doubleValue());
    }

    private static Number fncMod(Number arg0, Number arg1)
    {
        if ((arg0 instanceof Long) && (arg1 instanceof Long))
            return Long.valueOf(arg0.longValue() % arg1.longValue());
        else
            return Double.valueOf(arg0.doubleValue() % arg1.doubleValue());
    }

    private static Number fncSin(Number arg)
    {
        return Double.valueOf(Math.sin(arg.doubleValue()));
    }

    private static Number fncCos(Number arg)
    {
        return Double.valueOf(Math.cos(arg.doubleValue()));
    }

    private static Number fncTan(Number arg)
    {
        return Double.valueOf(Math.tan(arg.doubleValue()));
    }

    private static Number fncTanh(Number arg)
    {
        return Double.valueOf(Math.tanh(arg.doubleValue()));
    }

    private static Number fncAtan(Number arg)
    {
        return Double.valueOf(Math.atan(arg.doubleValue()));
    }

    private static Number fncAsin(Number arg)
    {
        return Double.valueOf(Math.asin(arg.doubleValue()));
    }

    private static Number fncAcos(Number arg)
    {
        return Double.valueOf(Math.acos(arg.doubleValue()));
    }

    private static Number fncExp(Number arg)
    {
        return Double.valueOf(Math.exp(arg.doubleValue()));
    }

    private static Number fncLn(Number arg)
    {
        return Double.valueOf(Math.log(arg.doubleValue()));
    }

    private static Number fncLog10(Number arg)
    {
        return Double.valueOf(Math.log10(arg.doubleValue()));
    }

    private static Number fncSqrt(Number arg)
    {
        return Double.valueOf(Math.sqrt(arg.doubleValue()));
    }

    private static Number fncCbrt(Number arg)
    {
        return Double.valueOf(Math.cbrt(arg.doubleValue()));
    }

    private static Number fncCosh(Number arg)
    {
        return Double.valueOf(Math.cosh(arg.doubleValue()));
    }

    private static Number fncAbs(Number a)
    {
        if (a instanceof Long) {
            return Long.valueOf(Math.abs(a.longValue()));
        }
        else {
            return Double.valueOf(Math.abs(a.doubleValue()));
        }
    }

    private static Number fncPower(Number base, Number exp)
    {
        return Double.valueOf(Math.pow(base.doubleValue(), exp.doubleValue()));
    }

    private static Number fncHypot(Number a, Number b)
    {
        return Double.valueOf(Math.hypot(a.doubleValue(), b.doubleValue()));
    }

    private static Number fncCmp(Number a, Number b)
    {
        if ((a instanceof Long) && (b instanceof Long)) {
            if (a.longValue() == b.longValue())
                return Long.valueOf(1);
            else
                return Long.valueOf(0);
        }
        else
        // 10 ulps should be enough for not very complicated expressions
        if (Math.abs(a.doubleValue() - b.doubleValue()) > (10 * Math.ulp(a
                .doubleValue())))
            return Long.valueOf(0);
        else
            return Long.valueOf(1);
    }

    private static Number fncGT(Number a, Number b)
    {
        if ((a instanceof Long) && (b instanceof Long)) {
            if (a.longValue() > b.longValue())
                return Long.valueOf(1);
            else
                return Long.valueOf(0);
        }
        else
        // 10 ulps should be enough for not very complicated expressions
        if (a.doubleValue() > (b.doubleValue() + 10 * Math.ulp(a.doubleValue())))
            return Long.valueOf(1);
        else
            return Long.valueOf(0);
    }

    private static Number fncGTE(Number a, Number b)
    {
        if ((a instanceof Long) && (b instanceof Long)) {
            if (a.longValue() >= b.longValue())
                return Long.valueOf(1);
            else
                return Long.valueOf(0);
        }
        else
        // 10 ulps should be enough for not very complicated expressions
        if ((a.doubleValue() + 10 * Math.ulp(a.doubleValue())) >= b
                .doubleValue())
            return Long.valueOf(1);
        else
            return Long.valueOf(0);
    }

    /**
     * Parses string expresion.
     * <p>
     * This method is provided for other macros.
     * </p>
     * 
     * @return result of evaluation
     * @throws java.lang.IllegalArgumentException
     */
    Number parse(String arg) throws IllegalArgumentException
    {
        exp = Tools.removeSpaces(arg);

        if (exp.length() == 0)
            throw new IllegalArgumentException("Macro Eval: missing expression");

        return parse(EndType.EXPR, Priority.PARENTHESIS);
    }

    /**
     * Parses string (sub-)expresion.
     * 
     * @param prev
     *            priority of previous operation
     * @return result of evaluation
     * @throws java.lang.IllegalArgumentException
     */
    private Number parse(EndType end, Priority prior)
            throws IllegalArgumentException
    {
        Number tmpres;
        Token t = nextToken();

        if ((end == EndType.PARAM) && (t.isOp(")"))) {
            // exp = exp.substring(1);
            return NO_MORE_PARAMS;
        }

        boolean sign = false;
        if (t.isOp("-")) {
            sign = true;
            t = nextToken();
        }// if

        if (t.isOp("(")) {
            tmpres = parse(EndType.BRACKET, Priority.PARENTHESIS);
            exp = exp.substring(1);
        }
        else if (t.isNumber()) {
            tmpres = t.asNumber();
        }
        else if (t.type == Token.Type.ID) {
            tmpres = evalFunction(t);
        }
        else {
            throw new IllegalArgumentException("Unexpected token");
        }

        if (sign) tmpres = fncNeg(tmpres);

        while (true) {

            if (exp.length() == 0) {
                if (end == EndType.EXPR)
                    return tmpres;
                else
                    throw new IllegalArgumentException(
                            "Unexpected end of expression");
            }// if

            char op = exp.charAt(0);

            switch (op) {
            case ')':
                if ((end == EndType.PARAM) || (end == EndType.BRACKET)) {
                    return tmpres;
                }
                else
                    throw new IllegalArgumentException(
                            "Unexpected character: \")\"");
            case ',':
                if (end == EndType.PARAM) {
                    exp = exp.substring(1);
                    return tmpres;
                }
                else
                    throw new IllegalArgumentException(
                            "Unexpected character: \",\"");
            case '=':
                if (prior.getPriority() >= Priority.CMP.getPriority())
                    return tmpres;
                exp = exp.substring(1);
                tmpres = fncCmp(tmpres, parse(end, Priority.CMP));
                break;
            case '>':
                if (prior.getPriority() >= Priority.CMP.getPriority())
                    return tmpres;
                exp = exp.substring(1);
                tmpres = fncGT(tmpres, parse(end, Priority.CMP));
                break;
            case '<':
                if (prior.getPriority() >= Priority.CMP.getPriority())
                    return tmpres;
                exp = exp.substring(1);
                tmpres = fncGT(parse(end, Priority.CMP), tmpres);
                break;
            case '+':
                if (prior.getPriority() >= Priority.ADD.getPriority())
                    return tmpres;
                exp = exp.substring(1);
                tmpres = fncSum(tmpres, parse(end, Priority.ADD));
                break;
            case '-':
                if (prior.getPriority() >= Priority.ADD.getPriority())
                    return tmpres;
                exp = exp.substring(1);
                tmpres = fncSub(tmpres, parse(end, Priority.ADD));
                break;
            case '*':
                if (prior.getPriority() >= Priority.MUL.getPriority())
                    return tmpres;
                exp = exp.substring(1);
                tmpres = fncMul(tmpres, parse(end, Priority.MUL));
                break;
            case '/':
                if (prior.getPriority() >= Priority.MUL.getPriority())
                    return tmpres;
                exp = exp.substring(1);
                tmpres = fncRealDiv(tmpres, parse(end, Priority.MUL));
                break;
            case '^':
                exp = exp.substring(1);
                tmpres = fncPower(tmpres, parse(end, Priority.POWER));
                break;
            case '%':
                if (prior.getPriority() >= Priority.MUL.getPriority())
                    return tmpres;
                exp = exp.substring(1);
                tmpres = fncMod(tmpres, parse(end, Priority.MUL));
                break;
            }

        }// while

    }// parse

}
