package pmp.macro.str;

import java.security.SecureRandom;
import java.util.Random;

import pmp.AbstractMacro;
import pmp.Macroprocessor;

/**
 * Generates random UUID/GUID.
 * For generating random numbers is used
 * {@link java.security.SecureRandom} with algorithm <code>"SHA1PRNG"</code>.
 * If not available, then standard {@link java.util.Random} class is used.
 * <p>Title: PMP: Macroprocessor</p>
 * <p>Description: Java macroprocessor</p>
 * <p>Copyright: Copyright (c) 2005</p>

 * @author Lud\u011Bk Hlav\u00E1\u010Dek
 * @version 1.0
 * @see java.util.UUID
 */

public class UUID
    extends AbstractMacro {

  private static final char[] digits = {
      '0', '1', '2', '3', '4', '5', '6', '7',
      '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'};

  private static void appendHexByte(int b, StringBuilder str)
  {
    int bh = ( (b >>> 4) & 0x0F);
    int bl = (b & 0x0F);
    str.append(digits[bh]).append(digits[bl]);
  }

  private static Random random;

  public UUID()
  {
    if (random == null) {
      long seed = System.currentTimeMillis() ^ (new Object()).hashCode()
          ^ (Runtime.getRuntime().freeMemory());
      try {
        String alg = "SHA1PRNG";

        // *** direct import of SecureRandom  ***
        random = SecureRandom.getInstance(alg);

        // *** load using java.lang.Class ***
        //Class randomClass = Class.forName("java.security.SecureRandom");
        //java.lang.reflect.Method getInstance = randomClass.getMethod("getInstance", new Class[] {alg.getClass()});
        //random = (Random) getInstance.invoke(null, new String[] {alg});
      }
      catch (Exception ex) {
        // loading of secure random failed - but we still have regular Random
         random = new Random(seed);
      } //catch
    } //if
  }

  public static final int PARAMS_REQUIRED = 1;

  /** Returns new UUID without hyphens. */
  public static final String COMPACT = "COMPACT";
  /** Returns new UUID in Windows registry format (ie: surrounded with curly brackets). */
  public static final String REGISTRY = "REGISTRY";
  /** Returns new UUID with <code>"uuid:"</code> prefix  */
  public static final String UUID_SCHEME = "UUID_SCHEME";

  private static final int MODE_DEFAULT = 0;
  private static final int MODE_COMPACT = 1;
  private static final int MODE_REGISTRY = 2;
  private static final int MODE_UUID_SCHEME = 3;

  /**
   * Generates new UUID.
   * This method creates new instance of this class and calls
   * {@link #run(Macroprocessor,String[])} method.
   * @param format format of generated uuid
   * @return generated UUID
   */
  public static String generate(String format)
  {
    int mode = MODE_DEFAULT;

    if ("".equals(format)) {
      mode = MODE_DEFAULT;
    }
    else if (COMPACT.equals(format)) {
      mode = MODE_COMPACT;
    }
    else if (REGISTRY.equals(format)) {
      mode = MODE_REGISTRY;
    }
    else if (UUID_SCHEME.equals(format)) {
      mode = MODE_UUID_SCHEME;
    }

    byte[] newUuid = new byte[16];

    random.nextBytes(newUuid); // something failed, refill buffer using regular Random

    newUuid[6] = (byte) (0x40 | (newUuid[6] & 0x0F)); // version = 0x4. [random]
    newUuid[8] = (byte) (0x80 | (newUuid[8] & 0x3F)); // variant = 10......

    StringBuilder str = new StringBuilder(48);
    if (mode == MODE_UUID_SCHEME) {
      str.append("uuid:");
    }
    if (mode == MODE_REGISTRY) {
      str.append("{");
    }

    for (int i = 0; i < 16; i++) {
      if (mode != MODE_COMPACT) {
        if ( (i == 4) || (i == 6) || (i == 8) || (i == 10)) {
          str.append('-');
        } //if
      } //if
      appendHexByte(newUuid[i], str);
    } //for

    if (mode == MODE_REGISTRY) {
      str.append("}");
    }

    return str.toString();

  }

  @Override
  public String run(Macroprocessor mp, String[] params)
  {
    return generate(params[1].trim().toUpperCase());
  }

}
