package pmp.xml;

import java.io.*;
import java.util.*;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.*;

import pmp.*;
import static pmp.xml.XMLTools.*;

/**
 * Loads XML configuration files.
 * <p>
 * Title: PMP: Macroprocessor
 * </p>
 * <p>
 * Description: Java macroprocessor
 * </p>
 * <p>
 * Copyright: Copyright (c) 2005
 * </p>
 * 
 * @author Lud\u011Bk Hlav\u00E1\u010Dek
 * @version 1.0
 */

public class XMLConfig extends Macroprocessor.Component {

    /** Location of default configuration file. */
    public static final String DEFAULT_CONFIG = "/pmp/config/default.xml";

    private final Map<String, MacroListItem> macros = new HashMap<String, MacroListItem>(
            64);

    /**
     * Reads configuration from default configuration file bundled with
     * macroprocessor.
     * 
     * @param mp
     *            instance of macroprocessor
     */
    public XMLConfig(Macroprocessor mp, boolean loaddef)
    {
        mp.super();

        if (loaddef) {
            InputStream is = XMLConfig.class
                    .getResourceAsStream(DEFAULT_CONFIG);
            if (is == null) { throw new FatalParserException(
                    "Couldn't load resource: " + DEFAULT_CONFIG); } // if
            loadConfigFile(is);
        }
    }

    /**
     * Reads configuration from file.
     * 
     * @param mp
     *            instance of macroprocessor
     * @param xmlConfig
     *            name of configuration file
     * @throws FatalParserException
     *             if file opening of file fails
     */
    public XMLConfig(Macroprocessor mp, String xmlConfig)
            throws FatalParserException
    {
        mp.super();
        try {
            InputStream is = new FileInputStream(xmlConfig);
            if (is == null) { throw new FatalParserException(
                    "Couldn't load file: " + xmlConfig); } // if
            loadConfigFile(is);
        } catch (Exception e) {
            throw new FatalParserException("Invalid configuration file "
                    + xmlConfig + ": " + e.getMessage());
        }
    }

    /**
     * Loads XML document from resource.
     * 
     * @param resourceName
     *            name of file
     * @return Document <code>true</code> if successful
     * @throws FatalParserException
     *             if something fails
     */
    private boolean loadConfigFile(InputStream resourceStream)
            throws FatalParserException
    {

        Document xmldoc = null;
        try {
            DocumentBuilderFactory factory = DocumentBuilderFactory
                    .newInstance();
            factory.setValidating(false);
            factory.setCoalescing(true);
            factory.setIgnoringComments(true);
            factory.setNamespaceAware(false);
            DocumentBuilder builder = factory.newDocumentBuilder();
            xmldoc = builder.parse(resourceStream);
            return parseXMLDocument(xmldoc);
        } // try
        catch (Exception ex) {
            throw new FatalParserException("Exception "
                    + ex.getClass().getName() + ": " + ex.getMessage());
        }
    }

    /**
     * Parses XML document.
     * 
     * @param doc
     *            DOM Document
     * @throws FatalParserException
     * @return <code>true</code> if successful
     */
    public boolean parseXMLDocument(Document doc) throws FatalParserException
    {

        Element root = doc.getDocumentElement();
        assert (ELEMENT_ROOT.equals(root.getTagName()));

        NodeList list = root.getChildNodes();

        for (int i = 0; i < list.getLength(); i++) {
            if (list.item(i).getNodeType() == Node.ELEMENT_NODE) {
                Element el = (Element) list.item(i);
                String name = el.getAttribute(ATTR_NAME);
                String elemText = getElementPCDATA(el);
                if (VALUE_BASE64.equals(el.getAttribute(ATTR_FORMAT))) {
                	throw new IllegalArgumentException("Base64 currently not supported.");
                    //elemText = Base64.decode(elemText);
                }
                boolean readOnly = "true".equalsIgnoreCase(el
                        .getAttribute(ATTR_READONLY));

                // macro definition
                if (ELEMENT_DEFINE.equals(el.getTagName())) {
                    AbstractMacro macro = new DefinedMacro(elemText);
                    if (el.hasAttribute(ATTR_OUTPUT_ENCODER)) {
                        String enc = el.getAttribute(ATTR_OUTPUT_ENCODER);
                        if (macros.containsKey(enc)) {
                            AbstractMacro oenc = macros.get(enc)
                                    .getDefinition();
                            macro = new MacroChain(macro, oenc);
                        }
                        else {
                            getMP().warning(
                                    "Invalid name of output-encoder macro.");
                        }// else
                    }// if
                    macros.put(name, new MacroListItem(macro, readOnly));
                }

                // classloading
                else if (ELEMENT_CLASS.equals(el.getTagName())) {
                    AbstractMacro macro = getMP().loadFromClass(elemText);
                    if (el.hasAttribute(ATTR_OUTPUT_ENCODER)) {
                        String enc = el.getAttribute(ATTR_OUTPUT_ENCODER);
                        if (macros.containsKey(enc)) {
                            AbstractMacro oenc = macros.get(enc)
                                    .getDefinition();
                            macro = new MacroChain(macro, oenc);
                        }
                        else {
                            getMP().warning(
                                    "Invalid name of output-encoder macro.");
                        }// else
                    }// if
                    macros.put(name, new MacroListItem(macro, readOnly));
                }

                else if (ELEMENT_UNDEF.equals(el.getTagName())) {
                    if (el.hasAttribute(ATTR_NAME)) {
                        macros.remove(el.getAttributeNode(ATTR_NAME));
                    }
                    else {
                        getMP()
                                .warning(
                                        "Attribute \"name\" missing at element\"undef\".");
                    }
                }

                else if (ELEMENT_IMPORT.equals(el.getTagName())) {
                    importConfig(elemText);
                }

                // external library
                else if (ELEMENT_LIBRARY.equals(el.getTagName())) {
                    getMP().importLibrary(elemText);
                }

            } // if
        } // for

        return true;
    }

    /**
     * Clears the content of lists of macros and projects
     */
    public void reset()
    {
        macros.clear();
    }

    /**
     * Imports configuration data from file given by <code>url</code> into
     * this configuration.
     * 
     * @param url
     *            location of source file
     */
    protected void importConfig(String url)
    {
        XMLConfig xml = new XMLConfig(getMP(), url);
        macros.putAll(xml.macros);
    }

    /**
     * Writes list of macros into given {@link MacroList}.
     * 
     * @param ml
     *            destination MacroList
     */
    public void writeToMacroList(MacroList ml)
    {
        Iterator<String> it = macros.keySet().iterator();
        while (it.hasNext()) {
            String key = it.next();
            MacroListItem item = macros.get(key);
            ml.setMacro(key, item.getDefinition());
        } // while
    }

}
