package pmp.xml;

import java.io.*;
import java.util.Arrays;
import java.util.Map;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import pmp.*;


/**
 * XML support routines.
 * <p>
 * Title: PMP: Macroprocessor
 * </p>
 * <p>
 * Description: Java macroprocessor
 * </p>
 * <p>
 * Copyright: Copyright (c) 2005
 * </p>
 * 
 * @author Lud\u011Bk Hlav\u00E1\u010Dek
 * @version 1.0
 */
public final class XMLTools {

    /**
     * Namespace. Not yet used.
     */
    public static final String NAMESPACE = "uuid:3f84b21f-56c1-4804-882d-ef1955036582";

    /**
     * Root element of configuration file.
     */
    public static final String ELEMENT_ROOT = "config";

    /**
     * Import another configuration file.
     */
    public static final String ELEMENT_IMPORT = "import";

    /**
     * Load library with class files.
     */
    public static final String ELEMENT_LIBRARY = "library";

    /**
     * Define macro.
     */
    public static final String ELEMENT_DEFINE = "define";

    /**
     * Load macro from given class file.
     */
    public static final String ELEMENT_CLASS = "class";

    /**
     * Common attribute for macro definitions.
     */
    public static final String ATTR_NAME = "name";

    /**
     * Common attribute for macro definitions. Specifies format of text in
     * current element.
     */
    public static final String ATTR_FORMAT = "format";

    /**
     * Base64 encoding.
     * 
     * @see #ATTR_FORMAT
     */
    public static final String VALUE_BASE64 = "base64";

    /**
     * No encoding.
     * 
     * @see #ATTR_FORMAT
     */
    public static final String VALUE_PLAIN = "plain";

    /**
     * Allow any further changes of this macro definitions.
     */
    public static final String ATTR_READONLY = "readOnly";

    public static final String ATTR_OUTPUT_ENCODER = "outputEncoder";

    public static final String ELEMENT_UNDEF = "undef";

    //public static final String ATTR_MASK = "mask";

    //public static final String ATTR_DEFAULT = "default";

    //public static final String ELEMENT_INPUT = "input";

    //public static final String ELEMENT_OUTPUT = "output";

    // public static final String ATTR_URL = "url";

    public static final String ATTR_ENCODING = "encoding";

    private static void writeln(OutputStream stream, String s)
            throws UnsupportedEncodingException, IOException
    {
        stream.write(s.getBytes("UTF-8"));
        stream.write("\r\n".getBytes("UTF-8"));
    }

    private static void write(OutputStream stream, String s)
            throws UnsupportedEncodingException, IOException
    {
        stream.write(s.getBytes("UTF-8"));
    }

    //private static Pattern p = Pattern.compile("[^ \t\r\n<>&]+");

    /**
     * Exports list of macros into XML file.
     * 
     * @param macros
     *            map with macros
     * @param fileName
     *            name of file
     * @return <code>null</code> if succeeded, otherwise error message
     */
    public static String exportConfig(Map<String, AbstractMacro> macros,
            String fileName)
    {
        try {
            final String h1 = "<?xml version=\"1.0\" encoding=\"utf-8\"?>\r\n";
            final String h2 = "<" + ELEMENT_ROOT + " xmlns=\"" + NAMESPACE
                    + "\">\r\n";

            FileOutputStream stream = new FileOutputStream(fileName);
            writeln(stream, h1);
            writeln(stream, h2);

            String[] keys = macros.keySet().toArray(new String[0]);
            Arrays.sort(keys);
            for (String macroName : keys) {
                AbstractMacro m = macros.get(macroName);
                String enc = "";

                if (m instanceof MacroChain) {
                    String encClass = ((MacroChain) m).getClass().getName();
                    enc = ATTR_OUTPUT_ENCODER + "=\"" + encClass + "\"";
                    m = ((MacroChain) m).getMacro();
                }

                if (m instanceof DefinedMacro) {
                    String ms = m.toString();
                    write(stream, "<" + ELEMENT_DEFINE);
                    write(stream, enc);
                    write(stream, " " + ATTR_NAME + "=\""
                            + xmlEncode(macroName) + "\">");
                    write(stream, xmlEncode(ms));
                    write(stream, "</" + ELEMENT_DEFINE + ">");
                }
                else {
                    write(stream, "<" + ELEMENT_CLASS + " ");
                    write(stream, enc);
                    write(stream, ATTR_NAME + "=\"" + xmlEncode(macroName)
                            + "\">");
                    write(stream, m.getClass().getName());
                    write(stream, "</" + ELEMENT_CLASS + ">");
                }
                writeln(stream, "");
            }// for
            writeln(stream, "</" + ELEMENT_ROOT + ">");
            stream.close();
        } catch (FileNotFoundException ex) {
            return "File not found.";
        } catch (IOException ex) {
            return "Write error: " + ex.getMessage();
        }
        return null;
    }

    /**
     * Returns text data from xml document element.
     * 
     * @param node
     *            element
     * @return text data from element
     */
    public static String getElementPCDATA(Element node)
    {
        node.normalize();
        NodeList list = node.getChildNodes();
        for (int i = 0; i < list.getLength(); i++) {
            Node child = list.item(i);
            if (child.getNodeType() == Node.TEXT_NODE) { return child
                    .getNodeValue(); } // if
        } // for
        return "";
    }

    /**
     * Encodes string to te be usable as value of element attribute or text
     * element.
     * <p>
     * Special characters &amp, &34;, &39;, &qt;, &lt; are converted to XML
     * entities.
     * </p>
     * 
     * @param str
     *            string to be encoded
     * @return encoded string.
     */
    public static String xmlEncode(String str)
    {

        StringBuilder result = new StringBuilder(str.length() * 2);

        for (int i = 0; i < str.length(); i++) {
            final char c = str.charAt(i);
            switch (c) {

            case '&':
                result.append("&amp;");
                break;
            case '\"':
                result.append("&#34;");
                break;
            case '\'':
                result.append("&#39;");
                break;
            case '<':
                result.append("&lt;");
                break;
            case '>':
                result.append("&gt;");
                break;
            default:
                result.append(c);
                break;
            } // switch
        } // for

        return result.toString();
    }

}
